import colorsys
import math
from collections import defaultdict
from enum import Enum
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import cast

import attr
import networkx as nx
import numpy as np
from godot_parser import GDSubResourceSection
from godot_parser import Node as GDNode
from matplotlib import patches as mpatches
from matplotlib import pyplot as plt
from matplotlib.axes._axes import Axes as MplAxes
from networkx import Graph
from scipy.spatial import KDTree
from scipy.spatial.transform import Rotation
from shapely import affinity
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.geometry import box as shapely_box
from trimesh import Trimesh
from trimesh import creation

from avalon.common.errors import SwitchError
from avalon.common.utils import only
from avalon.datagen.godot_base_types import FloatRange
from avalon.datagen.godot_base_types import IntRange
from avalon.datagen.world_creation.debug_plots import IS_DEBUG_VIS
from avalon.datagen.world_creation.geometry import Axis
from avalon.datagen.world_creation.geometry import BuildingTile
from avalon.datagen.world_creation.geometry import euclidean_distance
from avalon.datagen.world_creation.geometry import global_to_local_coords
from avalon.datagen.world_creation.geometry import local_to_global_coords
from avalon.datagen.world_creation.geometry import midpoint
from avalon.datagen.world_creation.indoor.blocks import BlocksByStory
from avalon.datagen.world_creation.indoor.blocks import CeilingBlock
from avalon.datagen.world_creation.indoor.blocks import FloorBlock
from avalon.datagen.world_creation.indoor.blocks import LadderBlock
from avalon.datagen.world_creation.indoor.blocks import LevelBlock
from avalon.datagen.world_creation.indoor.blocks import WallBlock
from avalon.datagen.world_creation.indoor.blocks import WindowBlock
from avalon.datagen.world_creation.indoor.blocks import make_blocks_from_tiles
from avalon.datagen.world_creation.indoor.blocks import new_make_block_nodes
from avalon.datagen.world_creation.indoor.components import Ladder
from avalon.datagen.world_creation.indoor.components import Ramp
from avalon.datagen.world_creation.indoor.components import Room
from avalon.datagen.world_creation.indoor.components import Story
from avalon.datagen.world_creation.indoor.components import StoryLink
from avalon.datagen.world_creation.indoor.components import Window
from avalon.datagen.world_creation.indoor.components import set_link_in_grid
from avalon.datagen.world_creation.indoor.components import set_room_tiles_in_grid
from avalon.datagen.world_creation.indoor.constants import CEILING_THICKNESS
from avalon.datagen.world_creation.indoor.constants import DEFAULT_FLOOR_THICKNESS
from avalon.datagen.world_creation.indoor.constants import EXPORT_MODE
from avalon.datagen.world_creation.indoor.constants import HIGH_POLY
from avalon.datagen.world_creation.indoor.constants import LOW_POLY
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.constants import CornerType
from avalon.datagen.world_creation.indoor.constants import ExportMode
from avalon.datagen.world_creation.indoor.constants import TileIdentity
from avalon.datagen.world_creation.indoor.helpers import rotate_position
from avalon.datagen.world_creation.indoor.mesh import MeshData
from avalon.datagen.world_creation.indoor.mesh import homogeneous_transform_matrix
from avalon.datagen.world_creation.indoor.mesh import make_color_visuals
from avalon.datagen.world_creation.indoor.mesh import unnormalize
from avalon.datagen.world_creation.indoor.tiles import draw_line_in_grid
from avalon.datagen.world_creation.indoor.tiles import find_corners
from avalon.datagen.world_creation.indoor.tiles import find_exterior_wall_footprints
from avalon.datagen.world_creation.indoor.utils import get_evenly_spaced_centroids
from avalon.datagen.world_creation.indoor.utils import inset_borders
from avalon.datagen.world_creation.region import Region
from avalon.datagen.world_creation.types import BuildingBoolNP
from avalon.datagen.world_creation.types import GDLinkedSection
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.types import RGBATuple
from avalon.datagen.world_creation.types import RawGDObject
from avalon.datagen.world_creation.utils import ARRAY_MESH_TEMPLATE
from avalon.datagen.world_creation.utils import make_transform


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class BuildingAestheticsConfig:
    rail_height: float = 0.75
    rail_thickness: float = 0.35
    rail_overhang_factor: float = 0.8  # 0-1 scalar of how much the building-top rail should hang over the edge

    window_height: float = 2
    window_width: float = 1.5
    window_min_gap: float = 1
    window_y_proportion_of_height = 0.55  # where to place window centroid, expressed as % of story height
    max_windows_per_wall: int = 8

    crossbeam_size: float = 0.35
    crossbeam_protrusion: float = 0.75
    crossbeam_min_gap: float = 2.5
    crossbeam_y_proportion_of_height: float = 0.9  # where to place beam centroid, expressed as % of story height

    crossbeam_color: RGBATuple = (0.19462, 0.13843, 0.02624, 1.0)
    exterior_color: RGBATuple = (0.66539, 0.56471, 0.28744, 1.0)
    trim_color: RGBATuple = (0.88792, 0.78354, 0.50289, 1.0)
    interior_wall_color: RGBATuple = (0.66539, 0.56471, 0.28744, 1.0)
    interior_wall_color_jitter: float = 0.02
    interior_wall_brightness_jitter: float = 0.025
    floor_color: RGBATuple = (0.88792, 0.78354, 0.50289, 1.0)
    ceiling_color: RGBATuple = (0.88792, 0.78354, 0.50289, 1.0)
    ladder_color: RGBATuple = (0.223529, 0.196078, 0.141176, 1.0)

    block_downsize_epsilon: float = 1e-5  # By how much to reduce block extents to avoid coincident faces


LevelBlockType = TypeVar("LevelBlockType", bound=LevelBlock)


@attr.s(auto_attribs=True, eq=False, collect_by_mro=True)
class Building:
    id: int
    stories: List[Story]
    story_links: List[StoryLink]
    position: Point3DNP
    yaw_degrees: float = 0
    is_climbable: bool = False
    aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig()

    def __attrs_post_init__(self) -> None:
        assert self.yaw_degrees == 0, "Building rotation currently not working, yaw_degrees must equal 0"

    @property
    def region(self) -> Region:
        return Region(
            x=FloatRange(self.offset_point[0], self.offset_point[0] + self.width),
            z=FloatRange(self.offset_point[2], self.offset_point[2] + self.length),
        )

    @property
    def height(self) -> FloatRange:
        return FloatRange(self.position[1], self.position[1] + sum([s.outer_height for s in self.stories]))

    @property
    def width(self) -> int:
        """
        Building width and height refers to its "bounding box"; the actual structure width/height is dependent
        on the footprints of individual stories. All story widths/heights are equal to each other and to the
        building width/height - variation in effective (traversable) width/height is all done via footprints.
        """
        return self.stories[0].width

    @property
    def length(self) -> int:
        return self.stories[0].length

    @property
    def offset_point(self) -> Point3DNP:
        # Where to offset any building space coordinates from to get the global ones
        return cast(Point3DNP, self.position - np.array([self.width / 2, 0, self.length / 2]))

    def with_transform(
        self, new_position: Optional[Point3DNP] = None, new_yaw_degrees: Optional[float] = None
    ) -> "Building":
        kwargs: Dict[str, Any] = {}
        if new_position is not None:
            kwargs["position"] = new_position
        if new_yaw_degrees is not None:
            kwargs["yaw_degrees"] = new_yaw_degrees
        return attr.evolve(self, **kwargs)

    def get_positive_height_at_point(self, point: Point3DNP) -> float:
        """
        Positive height refers to the exterior building height from it's y origin.
        Note: param point must be in building space.
        Returns np.nan if point lies outside building.
        """
        tile = math.floor(point[0]), math.floor(point[2])
        positive_height_at_point = 0.0
        for story in self.stories:
            out_of_bounds = tile[0] < 0 or tile[0] > story.width - 1 or tile[1] < 0 or tile[1] > story.length - 1
            if out_of_bounds:
                if story.num == 0:
                    return np.nan
                else:
                    continue
            elif story.footprint[tile[1], tile[0]]:
                positive_height_at_point += story.outer_height
                if story.num != 0:
                    # Negative terrain cannot be underground for non-ground floors.
                    positive_height_at_point += story.floor_negative_depth
        return positive_height_at_point

    def get_footprint_outline(self, story_num: int = 0) -> Polygon:
        tile_size = 1
        story = self.stories[story_num]
        corners = find_corners(story.footprint)
        points = []

        # Get corner centroids (positions are their top-left corners)
        for local_position, corner_type, is_outside in corners:
            # Note #1: Shapely ignores the Z coordinate, so we must make sure to only pass two coordinates here
            # We pass them as X -> X and Z -> Y
            # Note #2: Although this footprint is in building space, we _do_ have to offset by negative half-width since
            # the building origin (0,0) in Godot is its centroid and not it's top left-left corner as it is for tiles.
            points.append(
                Point(
                    local_position[1] - self.width / 2 + tile_size / 2,
                    local_position[0] - self.length / 2 + tile_size / 2,
                )
            )
        outline = Polygon(points)

        # Scale polygon to match OUTER edge of self, not tile midline
        min_x, min_y, max_x, max_y = outline.bounds
        width, length = max_x - min_x, max_y - min_y
        tile_size = 1
        x_scaling_factor = (width + tile_size) / width
        z_scaling_factor = (length + tile_size) / length
        outline = affinity.scale(outline, x_scaling_factor, z_scaling_factor)
        return outline

    def translate_block_to_building_space(self, block: LevelBlock, story: Story, epsilon: float = 0) -> LevelBlock:
        x_offset = -story.width / 2
        y_offset = self.get_story_y_offset(story.num)
        z_offset = -story.length / 2

        width, height, length = block.size
        adjusted_size = np.array([width - epsilon, height - epsilon, length - epsilon])
        x = x_offset + block.x.min_ge + width / 2
        y = y_offset + block.y.min_ge + height / 2
        z = z_offset + block.z.min_ge + length / 2
        centroid = np.array([x, y, z])
        return type(block).make(centroid, adjusted_size, block.rotation)

    def _get_pretty_visual_nodes(
        self,
        scene: GodotScene,
        blocks_by_story_by_kind: BlocksByStory,
        building_material: GDLinkedSection,
        building_name: str,
    ) -> List[GDNode]:
        DEFAULT_MESH = HIGH_POLY
        meshes = make_pretty_building_meshes(self, blocks_by_story_by_kind)
        visual_nodes = []
        for mesh_name, mesh in meshes.items():

            def _print_array(array: np.ndarray, precision: int) -> str:
                return ", ".join([f"{x:.{precision}f}" for x in array.flatten()])

            mesh_kwargs = {}
            colors = mesh.visual.vertex_colors / 255
            mesh_kwargs[f"surfaces/0"] = RawGDObject(
                ARRAY_MESH_TEMPLATE.format(
                    aabb=_print_array(mesh.bounds, 3),
                    vertex_floats=_print_array(mesh.vertices, 3),
                    vertex_normal_floats=_print_array(mesh.vertex_normals, 3),
                    color_floats=_print_array(colors, 5),
                    triangle_indices=_print_array(mesh.faces, 0),
                    index_count=str(len(mesh.faces) * 3),
                    vertex_count=str(len(mesh.vertices) * 3),
                    material_resource_type="SubResource"
                    if isinstance(building_material, GDSubResourceSection)
                    else "ExtResource",
                    material_id=building_material.id,
                    mesh_name=building_name,
                )
            )
            building_mesh = scene.add_sub_resource("ArrayMesh", resource_name="terrain", **mesh_kwargs)
            is_visible = mesh_name == DEFAULT_MESH
            mesh_node = GDNode(
                mesh_name,
                type="MeshInstance",
                properties={"mesh": building_mesh.reference, "material/0": "null", "visible": is_visible},
            )
            visual_nodes.append(mesh_node)
        return visual_nodes

    def export(
        self,
        scene: GodotScene,
        parent_node: GDNode,
        building_name: str,
        is_indoor_lighting_enabled: bool = False,
        scaling_factor: float = 1,
    ) -> None:
        blocks_by_story: BlocksByStory = {}
        for story in self.stories:
            blocks_by_story[story.num] = self._generate_blocks(story.num)

        building_material = scene.add_sub_resource("SpatialMaterial", vertex_color_use_as_albedo=True, roughness=0.5)
        building_script = scene.add_ext_resource("res://entities/building.gd", "Script")

        rotation = Rotation.from_euler("y", self.yaw_degrees, degrees=True).as_matrix().flatten()
        building_node = GDNode(
            building_name,
            type="Spatial",
            properties={
                "transform": make_transform(position=self.position, rotation=rotation),
                "script": building_script.reference,
                "is_climbable": self.is_climbable,
            },
        )
        parent_node.add_child(building_node)

        mesh_group = GDNode("meshes", type="Spatial")
        building_node.add_child(mesh_group)

        if EXPORT_MODE == ExportMode.PRETTY:
            for visual_node in self._get_pretty_visual_nodes(scene, blocks_by_story, building_material, building_name):
                mesh_group.add_child(visual_node)

        light_group = GDNode("lights", type="Spatial")
        building_node.add_child(light_group)
        if is_indoor_lighting_enabled:
            building_y_midpoint = (self.height.size / 2) - DEFAULT_FLOOR_THICKNESS
            light_centroid = 0, building_y_midpoint, 0
            first_story = self.stories[0]
            indoor_light = GDNode(
                f"OmniLight{self.id}",
                type="OmniLight",
                properties={
                    "transform": make_transform(position=light_centroid),
                    "light_energy": 0.5,
                    "omni_range": max(
                        math.sqrt(first_story.width**2 + first_story.length**2) / 2, self.height.size / 2
                    ),
                },
            )
            light_group.add_child(indoor_light)

        static_body_group = GDNode("static_bodies", type="Spatial")
        building_node.add_child(static_body_group)

        for story_num, blocks in blocks_by_story.items():
            story = self.stories[story_num]

            for i, block in enumerate(blocks):
                if not block.is_collider:
                    # When generating meshes, we only care for collider blocks here
                    continue

                block_in_building_space = self.translate_block_to_building_space(block, story)
                block_script = parent_extra_props = None
                if isinstance(block, FloorBlock) and block.is_climbable:
                    block_script = scene.add_ext_resource("res://items/climbable.gd", "Script").reference
                    parent_extra_props = dict(entity_id=0)
                story_block_nodes = new_make_block_nodes(
                    scene,
                    position=np.array(block_in_building_space.centroid),
                    size=np.array(block_in_building_space.size),
                    rotation=block_in_building_space.rotation.as_matrix().flatten(),
                    make_mesh=EXPORT_MODE == ExportMode.DEBUG and block.is_visual,
                    mesh_material=building_material,
                    parent_name=f"story_{story_num}_{block.type_name}_static_body_{i}",
                    parent_script=block_script,
                    parent_extra_props=parent_extra_props,
                )
                for block_node in story_block_nodes:
                    static_body_group.add_child(block_node)

        story_link_material = scene.add_ext_resource("res://shaders/BasicColor.material", "Material")
        story_links_node = GDNode(f"story_links", type="Spatial")
        building_node.add_child(story_links_node)
        include_meshes = EXPORT_MODE == ExportMode.DEBUG

        for story_link_num, story_link in enumerate(self.story_links):
            bottom_story = self.stories[story_link.bottom_story_id]
            top_story = self.stories[story_link.top_story_id]
            story_link_blocks = story_link.get_level_blocks(bottom_story, top_story)
            for i, level_block in enumerate(story_link_blocks):
                translated_block = self.translate_block_to_building_space(level_block, bottom_story)
                # Story links need to be additionally offset by the floor thickness
                translated_block.y = FloatRange(
                    translated_block.y.min_ge + DEFAULT_FLOOR_THICKNESS,
                    translated_block.y.max_lt + DEFAULT_FLOOR_THICKNESS,
                )

                parent_script = None
                if isinstance(level_block, (WallBlock, LadderBlock)) and level_block.is_climbable:
                    parent_script = scene.add_ext_resource("res://items/climbable.gd", "Script").reference

                story_link_block_nodes = new_make_block_nodes(
                    scene,
                    position=np.array(translated_block.centroid),
                    size=np.array(translated_block.size),
                    rotation=translated_block.rotation.as_matrix().flatten(),
                    make_mesh=include_meshes,
                    mesh_material=story_link_material,
                    parent_name=f"story_link_{story_link_num}_{i}",
                    parent_script=parent_script,
                    parent_extra_props=dict(entity_id=0),
                )
                for story_link_block_node in story_link_block_nodes:
                    story_links_node.add_child(story_link_block_node)

    def generate_tiles(
        self, story_num: int, include_links_going_up: bool = True, include_links_going_down: bool = True
    ) -> BuildingBoolNP:
        story = self.stories[story_num]
        return story.generate_tiles(include_links_going_up, include_links_going_down)

    def _generate_blocks(self, story_num: int) -> List[LevelBlock]:
        story = self.stories[story_num]
        return [
            *self._generate_terrain_blocks(story_num),
            *self._generate_hallway_floor_blocks(story_num),
            *self._generate_hallway_ceiling_blocks(story_num),
            *self._generate_interior_wall_blocks(story_num),
            *self._generate_exterior_wall_blocks(story_num),
            *(self._generate_ceiling_blocks(story_num) if story.has_ceiling else []),
            *self._generate_window_blocks(story_num),
        ]

    def get_story_y_offset(self, requested_story_num: int) -> float:
        # Returns the y position for where it's positive floor should begin
        assert requested_story_num <= len(self.stories)

        y: float = -DEFAULT_FLOOR_THICKNESS  # so that y=0 usually equates to top of ground story floor
        for i, story in enumerate(self.stories):
            if story.num > 0:
                y += story.floor_negative_depth
            if story.num == requested_story_num:
                return y
            y += story.outer_height
        raise ValueError("Requested story num was not hit")

    def _generate_interior_wall_blocks(self, story_num: int) -> Sequence[WallBlock]:
        story = self.stories[story_num]
        wall_blocks: List[WallBlock] = []
        story_tiles = self.generate_tiles(story_num)
        is_outline = story.get_outline_mask()
        is_wall = np.invert(story_tiles.astype(np.bool_))
        interior_wall_tiles = is_wall & ~is_outline
        interior_wall_blocks = make_blocks_from_tiles(interior_wall_tiles, solid_tile_value=True)

        for block in interior_wall_blocks:
            stretched_block = WallBlock(
                block.x,
                FloatRange(-story.floor_negative_depth, story.outer_height - CEILING_THICKNESS),
                block.z,
                is_interior=True,
            )
            cut_blocks = self._cut_windows(stretched_block, story.windows)
            wall_blocks.extend(cut_blocks)
        return wall_blocks

    def _generate_exterior_wall_blocks(self, story_num: int) -> Sequence[WallBlock]:
        story = self.stories[story_num]
        wall_blocks: List[WallBlock] = []
        story_tiles = self.generate_tiles(story_num)
        is_outline = story.get_outline_mask()

        exterior_wall_tiles = np.full_like(story_tiles, fill_value=TileIdentity.ROOM.value)
        exterior_wall_tiles[is_outline] = story_tiles[is_outline]
        exterior_wall_blocks = make_blocks_from_tiles(exterior_wall_tiles)

        for block in exterior_wall_blocks:
            offset_block = WallBlock(
                block.x, FloatRange(-story.floor_negative_depth, story.outer_height), block.z, is_interior=False
            )
            cut_blocks = self._cut_windows(offset_block, story.windows)
            wall_blocks.extend(cut_blocks)
        return wall_blocks

    def _generate_hallway_floor_blocks(self, story_num: int) -> Sequence[FloorBlock]:
        # Note that to cover all cases, we actually want to build each hallway separately, making staircases between
        # rooms of different heights. For now assume hallways only connect height=1(default floor thickness) terrain.
        story = self.stories[story_num]
        hallway_tiles = np.zeros((story.length, story.width))
        for hallway in story.hallways:
            draw_line_in_grid(hallway.points, hallway_tiles, TileIdentity.HALLWAY.value, include_ends=False)

        room_tiles = np.zeros((story.length, story.width))
        for room in story.rooms:
            set_room_tiles_in_grid(room, room.tiles, room_tiles)
        no_room_tiles = ~room_tiles.astype(np.bool_)

        # Ensure no hallway blocks are placed outside footprint (e.g. for entrance hallways leading to exterior
        # Also ensure we're only building hallway blocks outside rooms to avoid z-fighting
        hallway_tiles = (hallway_tiles.astype(np.bool_) & story.footprint & no_room_tiles).astype(np.float64)
        hallway_tiles[hallway_tiles == 1] = TileIdentity.HALLWAY.value

        return [
            FloorBlock(block.x, FloatRange(-story.floor_negative_depth, DEFAULT_FLOOR_THICKNESS), block.z)
            for block in make_blocks_from_tiles(hallway_tiles, solid_tile_value=TileIdentity.HALLWAY.value)
        ]

    def _generate_hallway_ceiling_blocks(self, story_num: int) -> Sequence[CeilingBlock]:
        """
        These blocks restrict the hallway height if is less than the room inner height.
        """
        story = self.stories[story_num]
        hallway_height_tiles = np.zeros((story.length, story.width), dtype=np.int8)
        for hallway in story.hallways:
            if hallway.height is not None:
                draw_line_in_grid(hallway.points, hallway_height_tiles, hallway.height, include_ends=False)

        room_tiles = np.zeros((story.length, story.width))
        for room in story.rooms:
            set_room_tiles_in_grid(room, room.tiles, room_tiles)
        no_room_mask = ~room_tiles.astype(np.bool_)

        # We don't want reduced height inside rooms or anything outside the building footprint
        valid_hallway_mask = (hallway_height_tiles.astype(np.bool_) & story.footprint & no_room_mask).astype(
            np.float64
        )
        hallway_height_tiles[valid_hallway_mask != 1] = 0
        unique_heights = np.unique(hallway_height_tiles[hallway_height_tiles != 0])

        blocks = []
        for height in unique_heights:
            height_marker = -1
            tiles_for_height = hallway_height_tiles.copy()
            tiles_for_height[tiles_for_height == height] = height_marker
            blocks.extend(
                [
                    CeilingBlock(block.x, FloatRange(DEFAULT_FLOOR_THICKNESS + height, story.outer_height), block.z)
                    for block in make_blocks_from_tiles(tiles_for_height, solid_tile_value=height_marker)
                ]
            )
        return blocks

    def _generate_terrain_blocks(self, story_num: int) -> Sequence[FloorBlock]:
        # "Terrain" for rooms and buildings works as follows:
        # 0 is ground level - a floor heightmap of all 0s equates to no floor blocks. Walls are built atop this level.
        # >=1 is positive terrain - blocks built inside the height of the room. A floor heightmap equating to wall
        #   height would end up in a non-existent room (floor touches ceiling).
        # <=-1 is chasms (gaps) downward into the floor, and only really make sense if you have some non-negative
        #   terrain. Walls blocks and positive terrain are all extended downwards to enclose the chasms.

        all_blocks: List[FloorBlock] = []
        story = self.stories[story_num]
        min_height = -story.floor_negative_depth

        ramp_cut_tiles = np.zeros_like(story.footprint, dtype=np.int_)
        for story_link in story.story_links:
            if story_link.top_story_id == story.num:
                set_landings = isinstance(story_link, Ladder)
                set_link_in_grid(
                    story_link, ramp_cut_tiles, set_bottom_landing=set_landings, set_top_landing=set_landings
                )
        ramp_cut_tiles = ramp_cut_tiles.astype(np.bool_)

        for room in story.rooms:
            # The floor heightmap gets overridden by cuts necessary in the floor to accommodate ramps leading up here
            tiles = room.floor_heightmap.copy()
            room_cut_tiles_array = ramp_cut_tiles[
                room.position.z : room.position.z + room.length, room.position.x : room.position.x + room.width
            ]
            room_cut_tiles = list(zip(*[list(a) for a in room_cut_tiles_array.nonzero()]))
            for z, x in room_cut_tiles:
                tiles[z, x] = min_height

            unique_heights = np.unique(tiles)
            for height in unique_heights:
                if height == min_height:
                    continue
                for is_climbable in [True, False]:
                    tiles_for_height_and_type = tiles.copy()
                    tiles_for_height_and_type[
                        (tiles_for_height_and_type == height) & (room.climbable_mask == is_climbable)
                    ] = np.inf
                    blocks = make_blocks_from_tiles(tiles_for_height_and_type, solid_tile_value=np.inf)
                    for block in blocks:
                        offset_block = FloorBlock(
                            x=FloatRange(block.x.min_ge + room.position.x, block.x.max_lt + room.position.x),
                            y=FloatRange(min_height, height),
                            z=FloatRange(block.z.min_ge + room.position.z, block.z.max_lt + room.position.z),
                            is_climbable=is_climbable,
                        )
                        # Yes, there *can* be windows in terrain blocks (for separating walls inside a room for example)
                        cut_blocks = self._cut_windows(offset_block, story.windows)
                        all_blocks.extend(cut_blocks)
        return all_blocks

    def _cut_windows(self, parent_block: LevelBlockType, windows: List[Window]) -> Sequence[LevelBlockType]:
        window_blocks = []
        # TODO: improve efficiency - we shouldn't need to do a O(N^2) loop with these
        for window in windows:
            window_block_in_story_space = WindowBlock.make(window.position, window.size)
            if (
                window_block_in_story_space.x.overlap(parent_block.x)
                and window_block_in_story_space.y.overlap(parent_block.y)
                and window_block_in_story_space.z.overlap(parent_block.z)
            ):
                local_coords = global_to_local_coords(
                    window.position,
                    np.array([parent_block.x.min_ge, parent_block.y.min_ge, parent_block.z.min_ge]),
                )
                window_block_in_parent_block_space = WindowBlock.make(local_coords, window.size)
                window_blocks.append(window_block_in_parent_block_space)
        if len(window_blocks) > 0:
            return self._cut_blocks(parent_block, window_blocks)
        else:
            return [parent_block]

    def _cut_blocks(
        self, parent_block: LevelBlockType, child_blocks: Sequence[LevelBlock]
    ) -> Sequence[LevelBlockType]:
        """
        Cut child_blocks as holes in parent_block in 2D
        is_visual / is_collider properties in parent are maintained in cut blocks.
        NOTE: child block coordinates must be localized to start at parent block origin
        """
        cut_axis = None
        for child_block in child_blocks:
            if child_block.x.size >= parent_block.x.size:
                block_cut_axis = Axis.X
            elif child_block.y.size >= parent_block.y.size:
                block_cut_axis = Axis.Y
            elif child_block.z.size >= parent_block.z.size:
                block_cut_axis = Axis.Z
            else:
                raise ValueError("Child blocks need to meet or exceed parent block size in at least one dimension")
            if cut_axis is None:
                cut_axis = block_cut_axis
            elif cut_axis != block_cut_axis:
                raise NotImplementedError(
                    "3d cutting is not supported yet; all child blocks must cut on the same axis"
                )

        if cut_axis == Axis.X:
            first_axis, second_axis = Axis.Y, Axis.Z
        elif cut_axis == Axis.Y:
            first_axis, second_axis = Axis.X, Axis.Z
        elif cut_axis == Axis.Z:
            first_axis, second_axis = Axis.X, Axis.Y
        else:
            raise SwitchError(f"Unset / unknown cut axis: {cut_axis}")

        # "tail" represents the leftover of the original block that doesn't fit into the tile_size=1 grid
        # we add it back to the blocks that touch the right or bottom of the grid when building the blocks below
        first_axis_size = getattr(parent_block, first_axis.value).size
        first_axis_tail = math.ceil(first_axis_size) - first_axis_size
        second_axis_size = getattr(parent_block, second_axis.value).size
        second_axis_tail = math.ceil(second_axis_size) - second_axis_size

        mult = 3  # resolution_multiplier - by what factor are we scaling up the grid's resolution
        grid = np.zeros((math.ceil(first_axis_size) * mult, math.ceil(second_axis_size) * mult))
        for child_block in child_blocks:
            first_range = getattr(child_block, first_axis.value)
            second_range = getattr(child_block, second_axis.value)
            first_range_min = round(first_range.min_ge * mult)
            if first_range_min < 0:
                first_range_min = 0
            second_range_min = round(second_range.min_ge * mult)
            if second_range_min < 0:
                second_range_min = 0

            grid[
                first_range_min : round(first_range.max_lt * mult),
                second_range_min : round(second_range.max_lt * mult),
            ] = 1

        cut_blocks = make_blocks_from_tiles(grid)
        # NOTE: the axis naming here is tricky:
        # `generate_blocks` uses its own internal axis naming: first axis is Z, second axis is X.
        # The meaning of those axes is different (since our cut axis arbitrary), so we need to map these back to our
        # meaning of first/second axis.
        # e.g. if we're cutting on X, the primary axis is Y and secondary is Z
        # the blocks we get back from generate_blocks were tiled as (z, x), so the mapping back is
        # z -> Y and x -> Z

        resulting_blocks = []
        for block in cut_blocks:
            # cut axis is untouched - we just set to whatever the parent had
            resulting_block = attr.evolve(parent_block, x=IntRange(0, 0), y=IntRange(0, 0), z=IntRange(0, 0))
            cut_axis_range: IntRange = getattr(parent_block, cut_axis.value)
            setattr(resulting_block, cut_axis.value, cut_axis_range)

            # Note that we need to scale all child block axis ranges down by dividing by mult
            first_axis_range = block.z
            parent_first_axis_offset = getattr(parent_block, first_axis.value).min_ge
            end_offset = parent_first_axis_offset
            if (
                first_axis_range.max_lt / mult + end_offset
                == getattr(parent_block, first_axis.value).max_lt - first_axis_tail
            ):
                end_offset += first_axis_tail
            offset_block_range = IntRange(
                first_axis_range.min_ge / mult + parent_first_axis_offset,
                first_axis_range.max_lt / mult + end_offset,
            )
            setattr(resulting_block, first_axis.value, offset_block_range)

            second_axis_range = block.x
            parent_second_axis_offset = getattr(parent_block, second_axis.value).min_ge
            end_offset = parent_second_axis_offset
            if (
                second_axis_range.max_lt / mult + end_offset
                == getattr(parent_block, second_axis.value).max_lt - second_axis_tail
            ):
                end_offset += second_axis_tail
            offset_block_range = IntRange(
                second_axis_range.min_ge / mult + parent_second_axis_offset,
                second_axis_range.max_lt / mult + end_offset,
            )
            setattr(resulting_block, second_axis.value, offset_block_range)
            resulting_blocks.append(resulting_block)
        return resulting_blocks

    def _generate_ceiling_blocks(self, story_num: int) -> Sequence[CeilingBlock]:
        story = self.stories[story_num]
        footprint = inset_borders(story.footprint).astype(np.int8)
        for story_link in story.story_links:
            if story_link.bottom_story_id == story_num:
                set_top_landing = isinstance(story_link, Ladder)
                set_link_in_grid(story_link, footprint, set_bottom_landing=True, set_top_landing=set_top_landing)

        return [
            CeilingBlock(
                x=block.x, y=FloatRange(story.outer_height - CEILING_THICKNESS, story.outer_height), z=block.z
            )
            for block in make_blocks_from_tiles(footprint, solid_tile_value=1)
        ]

    def _generate_window_blocks(self, story_num: int) -> Sequence[WindowBlock]:
        level_blocks: List[WindowBlock] = []
        for window in self.stories[story_num].windows:
            level_blocks.append(WindowBlock.make(window.position, window.size, is_collider=not window.is_passable))
        return level_blocks

    def plot(self) -> None:
        import seaborn as sns

        vmin = min(tile_symbol.value for tile_symbol in TileIdentity)
        colors = {
            "gray": TileIdentity.FULL.value,
            "lightgreen": TileIdentity.ROOM.value,
            "blue": TileIdentity.LINK.value,
            "darkblue": TileIdentity.LINK_BOTTOM_LANDING.value,
            "lightblue": TileIdentity.LINK_TOP_LANDING.value,
            "lightgray": TileIdentity.HALLWAY.value,
            "#eee": TileIdentity.VOID.value,
        }
        cmap = sorted(colors, key=colors.__getitem__)

        fig, axes = plt.subplots(len(self.stories), 1, figsize=(5, 5 * len(self.stories)))
        for i, story in enumerate(self.stories):
            tiles = self.generate_tiles(story.num)
            ax = axes[i] if len(self.stories) > 1 else axes

            # Rooms with their heightmap
            continuous_tiles = tiles.copy()
            continuous_tiles[tiles != TileIdentity.ROOM.value] = np.nan
            max_height = 0
            for room in story.rooms:
                continuous_tiles[
                    room.position.z : room.position.z + room.length, room.position.x : room.position.x + room.width
                ] = room.floor_heightmap
                max_height = max(max_height, room.floor_heightmap.max())
            continuous_tiles[continuous_tiles > 0] /= max_height
            sns.heatmap(
                continuous_tiles,
                square=True,
                linewidths=0.25,
                linecolor="white",
                cmap="YlGn_r",
                vmin=-1,
                vmax=1,
                ax=ax,
            )

            # unset tiles, hallways, ramps
            discrete_tiles = tiles.copy()
            discrete_tiles[tiles == TileIdentity.ROOM.value] = np.nan
            sns.heatmap(
                discrete_tiles,
                square=True,
                linewidths=0.25,
                linecolor="white",
                cmap=cmap,
                cbar=False,
                vmin=vmin,
                vmax=0,
                ax=ax,
            )

            for r, room in enumerate(story.rooms):
                ax.annotate(str(r), (room.center.x + 0.25, room.center.z + 0.75))
            ax.title.set_text(f"Floor {story.num}")

        ax = axes[0] if len(self.stories) > 1 else axes
        handles, labels = ax.get_legend_handles_labels()
        manual_patches = [
            mpatches.Patch(color=color, label=TileIdentity(tile_value).pretty_name)
            for color, tile_value in colors.items()
        ]
        handles.extend(manual_patches)
        plt.legend(handles=handles, bbox_to_anchor=(1, 0), loc="lower right", bbox_transform=fig.transFigure, ncol=3)
        plt.show()

    def rebuild_rotated(self, yaw_degrees: int) -> "Building":
        if yaw_degrees == -180:
            yaw_degrees = 180
        assert yaw_degrees in {0, 90, -90, 180}
        numpy_rotation = int(-yaw_degrees / 90)

        new_story_links = []
        for story_link in self.story_links:
            bottom_story = self.stories[story_link.bottom_story_id]
            new_bottom_position = rotate_position(
                story_link.bottom_position, bottom_story.width, bottom_story.length, yaw_degrees
            )
            top_story = self.stories[story_link.top_story_id]
            new_top_position = rotate_position(story_link.top_position, top_story.width, top_story.length, yaw_degrees)
            new_story_links.append(
                attr.evolve(story_link, bottom_position=new_bottom_position, top_position=new_top_position)
            )

        rotated_stories = []
        for story in self.stories:
            if yaw_degrees in {90, -90}:
                new_story_width = story.length
                new_story_length = story.width
            else:
                new_story_width = story.width
                new_story_length = story.length

            new_footprint = np.rot90(story.footprint, numpy_rotation)

            new_rooms = []
            for room in story.rooms:
                if yaw_degrees in {90, -90}:
                    new_room_width = room.length
                    new_room_length = room.width
                else:
                    new_room_width = room.width
                    new_room_length = room.length

                new_position = rotate_position(room.position, story.width, story.length, yaw_degrees, tile_like=True)
                # Switch from rotated point to new top-left point
                if yaw_degrees == 90:
                    new_position.x -= room.length - 1
                elif yaw_degrees == -90:
                    new_position.z -= room.width - 1
                elif yaw_degrees == 180:
                    new_position.x -= room.width - 1
                    new_position.z -= room.length - 1
                assert new_position.x >= 0
                assert new_position.z >= 0

                new_heightmap = np.rot90(room.floor_heightmap, numpy_rotation)
                new_rooms.append(
                    attr.evolve(
                        room,
                        position=new_position,
                        width=new_room_width,
                        length=new_room_length,
                        floor_heightmap=new_heightmap,
                    )
                )

            new_hallways = []
            for hallway in story.hallways:
                new_points = tuple(
                    [rotate_position(point, story.width, story.length, yaw_degrees) for point in hallway.points]
                )
                new_hallways.append(attr.evolve(hallway, points=new_points))

            new_story_links_for_story = [
                story_link
                for story_link in new_story_links
                if story.num in {story_link.bottom_story_id, story_link.top_story_id}
            ]

            new_windows = []
            for window in story.windows:
                position_2d = BuildingTile(window.position[0], window.position[2])
                new_position_2d = rotate_position(position_2d, story.width, story.length, yaw_degrees, tile_like=False)
                new_position_3d = np.array([new_position_2d.x, window.position[1], new_position_2d.z])
                if yaw_degrees in {90, -90}:
                    new_size = np.array([window.size[2], window.size[1], window.size[0]])
                else:
                    new_size = window.size.copy()
                new_windows.append(attr.evolve(window, position=new_position_3d, size=new_size))

            new_entrances = []
            for entrance in story.entrances:
                new_angle = (entrance.azimuth.angle_from_positive_x + yaw_degrees) % 360
                new_azimuth = only([a for a in list(Azimuth) if a.angle_from_positive_x % 360 == new_angle])
                new_entrance_points = tuple(
                    [rotate_position(point, story.width, story.length, yaw_degrees) for point in entrance.points]
                )
                new_entrances.append(
                    attr.evolve(
                        entrance,
                        azimuth=new_azimuth,
                        points=new_entrance_points,
                    )
                )

            rotated_stories.append(
                attr.evolve(
                    story,
                    width=new_story_width,
                    length=new_story_length,
                    footprint=new_footprint,
                    rooms=new_rooms,
                    hallways=new_hallways,
                    story_links=new_story_links_for_story,
                    windows=new_windows,
                    entrances=new_entrances,
                )
            )
        return attr.evolve(self, stories=rotated_stories, story_links=new_story_links)


class BuildingNavGraph(Graph):
    def __init__(self, building: Building, excluded_stories: Tuple[int, ...] = tuple()) -> None:
        """
        Represents the building with rooms as nodes and hallways + story links as edges.
        Does not check for room traversability (there may be blocked hallways / links or impassable terrain) - use
        StoryNavGraph for these.
        """
        super(BuildingNavGraph, self).__init__()
        self.excluded_stories = excluded_stories
        self._add_story_edges(building)
        self._add_story_link_edges(building)

    def _add_story_edges(self, building: Building) -> None:
        y_offset = 0.0
        for story in building.stories:
            if story.num in self.excluded_stories:
                continue

            y_offset += story.floor_negative_depth
            for room in story.rooms:
                room_y = room.floor_heightmap[room.length // 2, room.width // 2]
                self.add_node(
                    self.get_room_node(story, room), position=(room.center.x, y_offset + room_y, room.center.z)
                )
            for i, hallway in enumerate(story.hallways):
                if hallway.from_room_id < 0 or hallway.to_room_id < 0:
                    # Exterior links
                    continue
                from_room = story.rooms[hallway.from_room_id]
                to_room = story.rooms[hallway.to_room_id]
                from_room_node = self.get_room_node(story, from_room)
                to_room_node = self.get_room_node(story, to_room)
                distance_to_hallway = euclidean_distance(from_room.center, hallway.points[0])
                distance_from_hallway = euclidean_distance(hallway.points[-1], to_room.center)
                total_length = hallway.total_length + distance_to_hallway + distance_from_hallway
                self.add_edge(from_room_node, to_room_node, distance=total_length)
            y_offset += story.outer_height

    def _add_story_link_edges(self, building: Building) -> None:
        for link in building.story_links:
            if link.bottom_story_id in self.excluded_stories or link.top_story_id in self.excluded_stories:
                continue
            assert isinstance(link, (Ramp, Ladder))
            from_story = building.stories[link.bottom_story_id]
            from_room = from_story.rooms[link.bottom_room_id]
            to_story = building.stories[link.top_story_id]
            to_room = to_story.rooms[link.top_room_id]
            from_room_node = self.get_room_node(from_story, from_room)
            to_room_node = self.get_room_node(to_story, to_room)
            distance_to_landing = euclidean_distance(from_room.center, link.bottom_position)
            distance_from_landing = euclidean_distance(link.top_position, to_room.center)
            total_length = link.get_link_length(from_story, to_story) + distance_to_landing + distance_from_landing
            self.add_edge(from_room_node, to_room_node, distance=total_length)

    def get_nearest_node(self, point: Point3DNP) -> str:
        positions: Dict[str, Tuple[int, int]] = nx.get_node_attributes(self, "position")
        kd_tree = KDTree(list(positions.values()))
        distance, position_idx = kd_tree.query(point)
        return cast(str, list(positions.keys())[position_idx])

    def get_room_node(self, story: Story, room: Room) -> str:
        return f"{story.num}_R{room.id}"

    def get_stories_and_rooms_by_node_id(self, stories: List[Story]) -> Dict[str, Tuple[Story, Room]]:
        stories_and_rooms_by_node_id = {}
        for story in stories:
            for room in story.rooms:
                stories_and_rooms_by_node_id[self.get_room_node(story, room)] = (story, room)
        return stories_and_rooms_by_node_id

    def plot(self, origin: Point3DNP = np.array([0, 0, 0]), ax: Optional[MplAxes] = None):
        plt.figure(figsize=(25, 25))
        positions = nx.get_node_attributes(self, "position")

        def translate_and_swap_axes(p: Point3DNP) -> Point3DNP:
            # This swap ensures the plot matches the building tile plot
            # to avoid axis swap, we need mpl 3.6.0+ to change up-axis: https://stackoverflow.com/a/56457693/5814943
            x, y, z = p
            x += origin[0]
            y += origin[1]
            z += origin[2]
            return np.array([z, x, y])

        node_xyz = np.array([translate_and_swap_axes(positions[v]) for v in self.nodes])
        edge_xyz = np.array(
            [(translate_and_swap_axes(positions[u]), translate_and_swap_axes(positions[v])) for u, v in self.edges()]
        )

        custom_ax = True
        if ax is None:
            custom_ax = False
            fig = plt.figure()
            fig.tight_layout()
            ax = fig.add_subplot(111, projection="3d")
        ax.scatter(*node_xyz.T, s=100, ec="w")
        for i, node in enumerate(self.nodes):
            ax.text(*node_xyz[i], node, zdir=None, size=8)

        distance_by_edge = nx.get_edge_attributes(self, "distance")
        for edge, edge_coordinates in zip(self.edges, edge_xyz):
            middle = midpoint(edge_coordinates[0], edge_coordinates[1])
            ax.plot(*edge_coordinates.T, color="tab:gray")
            ax.text(*middle, f"{distance_by_edge[edge]:.1f}", color="darkgray", size=7)

        if not custom_ax:
            zs = [z for x, y, z in node_xyz]
            ax.set_zlim(min(zs), max(zs) + 0.1)
            ax.set_xlabel("z")
            ax.set_ylabel("x")
            ax.set_zlabel("y")
        return ax


def get_building_footprint_outline(building: Building, story_num: int = 0) -> Polygon:
    """
    Returns a 2D outline (y is always 0) of specified story footprint outline in world space.
    NB! Z is mapped to Y in the polygon.
    """
    outline = building.get_footprint_outline(story_num)
    x_offset, _y_offset, z_offset = building.position
    outline = affinity.rotate(outline, -building.yaw_degrees)
    outline = affinity.translate(outline, xoff=x_offset, yoff=z_offset)
    return outline


def get_building_positive_height_at_point(point_in_world_space: Point3DNP, building: Building) -> float:
    """
    Positive height refers to the exterior building height from it's y origin.
    Returns np.nan if point is not within building footprint.
    NOTE: both point and building must be in world space (rotated and positioned)
    """
    outline = get_building_footprint_outline(building)
    if IS_DEBUG_VIS:
        x, z = zip(*outline.exterior.coords)
        ax = plt.axes()
        ax.plot(x, z)
        ax.plot(point_in_world_space[0], point_in_world_space[2], "ro")
        ax.invert_yaxis()
        plt.gca().set_aspect("equal")
        plt.show()
    point = Point(point_in_world_space[0], point_in_world_space[2])
    if not outline.contains(point):
        # Note: points ON the edge are not included
        return np.nan
    else:
        building_centroid_2d = building.position[0], building.position[2]
        unrotated_point = affinity.rotate(point, -building.yaw_degrees, origin=building_centroid_2d)
        point_in_building_space = global_to_local_coords(
            np.array([unrotated_point.x, 0, unrotated_point.y]), building.position
        )
        return building.get_positive_height_at_point(point_in_building_space)


class BuildingTask(Enum):
    EXPLORE = "EXPLORE"
    OPEN = "OPEN"
    PUSH = "PUSH"
    STACK = "STACK"
    CLIMB = "CLIMB"
    JUMP = "JUMP"


def get_block_color(block: LevelBlock, aesthetics: BuildingAestheticsConfig, rand: np.random.Generator) -> RGBATuple:
    if isinstance(block, WallBlock):
        if block.is_interior:
            color = aesthetics.interior_wall_color
            brightness_jitter = rand.normal(0.0, aesthetics.interior_wall_brightness_jitter)
            alpha = color[-1]
            h, s, v = colorsys.rgb_to_hsv(*color[:-1])
            v = float(np.clip(v + brightness_jitter, 0, 1))
            rgb = colorsys.hsv_to_rgb(h, s, v)
            rgb_jittered = tuple(channel + rand.normal(0.0, aesthetics.interior_wall_color_jitter) for channel in rgb)
            return (*rgb_jittered, alpha)  # type: ignore
        else:
            return aesthetics.exterior_color
    elif isinstance(block, FloorBlock):
        return aesthetics.floor_color
    elif isinstance(block, CeilingBlock):
        return aesthetics.ceiling_color
    elif isinstance(block, LadderBlock):
        return aesthetics.ladder_color
    else:
        raise SwitchError(f"Unknown block type {type(block)}")


def make_pretty_building_meshes(building: Building, blocks_by_story: BlocksByStory) -> Dict[str, Trimesh]:
    rand = np.random.default_rng(building.id)
    corners_by_story = defaultdict(list)
    wall_footprints_by_story = defaultdict(list)
    exterior_wall_blocks_by_story: DefaultDict[int, List[WallBlock]] = defaultdict(list)
    original_outlines_by_story = {}

    cornerless_building = building.rebuild_rotated(0)
    for story in cornerless_building.stories:
        outline = cornerless_building.get_footprint_outline(story.num)
        original_outlines_by_story[story.num] = outline

        new_footprint = story.footprint.copy()
        wall_footprints_by_story[story.num] = find_exterior_wall_footprints(new_footprint)
        corners = find_corners(new_footprint)
        corners_by_story[story.num] = corners
        for corner_position, corner_type, is_outside in corners:
            if is_outside:
                new_footprint[corner_position[0], corner_position[1]] = False
        story.footprint = new_footprint
        exterior_wall_blocks_by_story[story.num].extend(cornerless_building._generate_exterior_wall_blocks(story.num))

    mesh_datasets: DefaultDict[str, MeshData] = defaultdict(lambda: MeshData())

    x_offset = -building.width / 2
    z_offset = -building.length / 2
    tile_size = 1

    def add_mesh_data(parent_name: str, child_mesh: Trimesh) -> None:
        nonlocal mesh_datasets
        parent_mesh_data = mesh_datasets[parent_name]
        parent_mesh_data.vertices.extend(child_mesh.vertices)
        parent_mesh_data.faces.extend(child_mesh.faces + parent_mesh_data.index_offset)
        parent_mesh_data.face_normals.extend(child_mesh.face_normals)
        if child_mesh.visual:
            parent_mesh_data.face_colors.extend(child_mesh.visual.face_colors)
        else:
            parent_mesh_data.face_colors.extend([0.0, 0.0, 0.0] * len(child_mesh.faces))
        parent_mesh_data.index_offset += len(child_mesh.vertices)

    for story_num, blocks in blocks_by_story.items():
        story = building.stories[story_num]
        y_offset = building.get_story_y_offset(story_num)
        exterior_wall_blocks = exterior_wall_blocks_by_story[story.num]
        other_blocks = [block for block in blocks if not (isinstance(block, WallBlock) and not block.is_interior)]
        for block in [*exterior_wall_blocks, *other_blocks]:
            if not block.is_visual:
                continue

            translated_block = building.translate_block_to_building_space(
                block, story, building.aesthetics.block_downsize_epsilon
            )

            box = creation.box(translated_block.size)
            box = unnormalize(box)
            position = np.array([*translated_block.centroid])
            transform = homogeneous_transform_matrix(position)
            box.apply_transform(transform)
            box.visual = make_color_visuals(box, get_block_color(block, building.aesthetics, rand))
            add_mesh_data(HIGH_POLY, box)
            if (isinstance(block, WallBlock) and not block.is_interior) or isinstance(block, CeilingBlock):
                add_mesh_data(LOW_POLY, box)

        for story_link in story.story_links:
            if story_link.bottom_story_id != story.num:
                continue
            story_link_level_blocks = story_link.get_level_blocks(
                building.stories[story_link.bottom_story_id], building.stories[story_link.top_story_id]
            )
            for level_block in story_link_level_blocks:
                if not level_block.is_visual:
                    continue

                block_width, block_height, block_length = level_block.size
                block_x, block_y, block_z = level_block.centroid
                trimesh_box = creation.box((block_width, block_height, block_length))
                position = np.array(
                    [
                        x_offset + block_x,
                        building.get_story_y_offset(story_link.bottom_story_id) + DEFAULT_FLOOR_THICKNESS + block_y,
                        z_offset + block_z,
                    ]
                )
                rotation_transform = homogeneous_transform_matrix(rotation=level_block.rotation)
                trimesh_box.apply_transform(rotation_transform)
                translation_transform = homogeneous_transform_matrix(position)
                trimesh_box.apply_transform(translation_transform)
                trimesh_box = unnormalize(trimesh_box)
                trimesh_box.visual = make_color_visuals(
                    trimesh_box, get_block_color(level_block, building.aesthetics, rand)
                )
                add_mesh_data(HIGH_POLY, trimesh_box)

        outline = original_outlines_by_story[story_num]

        rail_height = building.aesthetics.rail_height
        rail_width = building.aesthetics.rail_thickness

        # constant: 1/2 of tile size
        buffer_width = 0.5

        min_x, min_y, max_x, max_y = outline.bounds
        width, length = max_x - min_x, max_y - min_y
        x_scaling_factor = width / (width + buffer_width * 2 + rail_width * building.aesthetics.rail_overhang_factor)
        z_scaling_factor = length / (length + buffer_width * 2 + rail_width * building.aesthetics.rail_overhang_factor)
        outline = outline.buffer(buffer_width, join_style=1, single_sided=True)
        outline = affinity.scale(outline, x_scaling_factor, z_scaling_factor)

        path = np.array([(c[0], y_offset + story.outer_height + rail_height, c[1]) for c in outline.exterior.coords])
        square = shapely_box(-rail_width / 2, 0, rail_width / 2, rail_height)
        border = creation.sweep_polygon(square, path)
        border = unnormalize(border)
        border.visual = make_color_visuals(border, building.aesthetics.trim_color)
        add_mesh_data(HIGH_POLY, border)

        offset = np.array([x_offset, 0, z_offset])

        crossbeam_size = building.aesthetics.crossbeam_size
        crossbeam_protrusion = building.aesthetics.crossbeam_protrusion
        crossbeam_centroid_y = y_offset + story.outer_height * building.aesthetics.crossbeam_y_proportion_of_height
        for wall_footprint in wall_footprints_by_story[story_num]:
            centroids = get_evenly_spaced_centroids(
                wall_footprint, crossbeam_size, building.aesthetics.crossbeam_min_gap, crossbeam_centroid_y
            )
            for centroid in centroids:
                centroid = local_to_global_coords(centroid, offset)
                transform = homogeneous_transform_matrix(position=centroid)
                if wall_footprint.is_vertical:
                    # Centroids are in wall middle, so adding +1 (wall thickness) to ensure they protrude
                    extents = (1 + crossbeam_protrusion, crossbeam_size, crossbeam_size)
                else:
                    extents = (crossbeam_size, crossbeam_size, 1 + crossbeam_protrusion)
                crossbeam = creation.box(extents, transform)
                crossbeam = unnormalize(crossbeam)
                crossbeam.visual = make_color_visuals(crossbeam, building.aesthetics.crossbeam_color)
                add_mesh_data(HIGH_POLY, crossbeam)

        for (z, x), corner_type, is_outside in corners_by_story[story_num]:
            if not is_outside:
                continue

            story = building.stories[story_num]

            corner_centroid = np.array(
                [x_offset + x + tile_size / 2, y_offset + (story.outer_height) / 2, z_offset + z + tile_size / 2]
            )
            rotation = Rotation.from_euler("x", 90, degrees=True)
            transform = homogeneous_transform_matrix(position=corner_centroid, rotation=rotation)
            round_corner_cylinder = creation.cylinder(tile_size / 2, story.outer_height, transform=transform)
            round_corner_cylinder = unnormalize(round_corner_cylinder)

            multiplier = 1 if corner_type in [CornerType.NE, CornerType.NW] else -1
            box_z = corner_centroid[2] + (multiplier * (tile_size / 4))
            box_centroid = np.array([corner_centroid[0], corner_centroid[1], box_z])
            transform = homogeneous_transform_matrix(position=box_centroid)
            horizontal_box = creation.box((tile_size, story.outer_height, tile_size / 2), transform)
            horizontal_box = unnormalize(horizontal_box)

            multiplier = 1 if corner_type in [CornerType.NW, CornerType.SW] else -1
            box_x = corner_centroid[0] + (multiplier * (tile_size / 4))
            box_centroid = np.array([box_x, corner_centroid[1], corner_centroid[2]])
            transform = homogeneous_transform_matrix(position=box_centroid)
            vertical_box = creation.box((tile_size / 2, story.outer_height, tile_size), transform)
            vertical_box = unnormalize(vertical_box)

            for component in [round_corner_cylinder, horizontal_box, vertical_box]:
                component.visual = make_color_visuals(component, building.aesthetics.exterior_color)
                add_mesh_data(HIGH_POLY, component)

    meshes = {}
    for name, mesh_data in mesh_datasets.items():
        mesh = Trimesh(
            mesh_data.vertices,
            mesh_data.faces,
            face_normals=mesh_data.face_normals,
            face_colors=mesh_data.face_colors,
            process=False,  # Don't merge identical vertices!
        )
        mesh.invert()  # Godot uses the opposite winding order of Trimesh
        meshes[name] = mesh
    return meshes



import math
from typing import List

import attr
import numpy as np
from godot_parser import Color
from godot_parser import Node as GDNode
from godot_parser import NodePath
from scipy.spatial.transform import Rotation

from avalon.common.utils import only
from avalon.datagen.world_creation.entities.doors.door import Door
from avalon.datagen.world_creation.entities.doors.locks.rotating_bar import RotatingBar
from avalon.datagen.world_creation.entities.doors.locks.sliding_bar import SlidingBar
from avalon.datagen.world_creation.entities.doors.types import HingeSide
from avalon.datagen.world_creation.entities.doors.types import LatchingMechanics
from avalon.datagen.world_creation.indoor.blocks import new_make_block_nodes
from avalon.datagen.world_creation.indoor.utils import get_scaled_mesh_node
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.utils import make_transform


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HingeDoor(Door):
    hinge_side: HingeSide = HingeSide.LEFT
    hinge_radius: float = 0.05
    max_inwards_angle: float = 90.0
    max_outwards_angle: float = 90.0
    handle_size_proportion: float = 0.2
    latching_mechanics: LatchingMechanics = LatchingMechanics.NO_LATCH

    def get_node(self, scene: GodotScene) -> GDNode:
        DOOR_BODY = "body"

        body_nodes = self._get_body_nodes(scene, DOOR_BODY)
        hinge_nodes = self._get_hinge_nodes(scene, DOOR_BODY)

        door_script = scene.add_ext_resource("res://entities/doors/hinge_door.gd", "Script")
        group_spatial_node = GDNode(
            f"hinge_door_{self.entity_id}",
            "Spatial",
            properties={
                "transform": make_transform(self.rotation, self.position),
                "script": door_script.reference,
                "entity_id": self.entity_id,
                "is_latched": self.latching_mechanics != LatchingMechanics.NO_LATCH,
            },
        )
        for body_node in body_nodes:
            group_spatial_node.add_child(body_node)
        for hinge_node in hinge_nodes:
            group_spatial_node.add_child(hinge_node)
        body_node = only(body_nodes)
        door_body_width = self.size[0]
        door_body_size = np.array([door_body_width, self.size[1], self.size[2]])
        body_x_offset = -door_body_width / 2 if self.hinge_side == HingeSide.RIGHT else door_body_width / 2
        for i, lock in enumerate(self.locks):
            # if we have two locks of the same kind, their instance names will otherwise crash
            lock_with_id = attr.evolve(lock, entity_id=i)
            group_spatial_node.add_child(lock_with_id.get_node(scene))
            offset_centroid = np.array([body_x_offset, 0, 0])
            for body_addition in lock_with_id.get_additional_door_body_nodes(scene, door_body_size, offset_centroid):
                body_node.add_child(body_addition)
        return group_spatial_node

    def _get_body_nodes(self, scene: GodotScene, door_body_node_name: str) -> List[GDNode]:
        # The door body here also includes the handle, since PinJoints suck in Godot, so we can't have them separate :/
        door_width, door_height, door_thickness = self.size

        if self.hinge_side == HingeSide.LEFT:
            body_offset = np.array([door_width / 2, 0, 0])
            body_position = np.array([-door_width / 2 + self.hinge_radius, 0, 0])
        else:
            body_offset = np.array([-door_width / 2, 0, 0])
            body_position = np.array([door_width / 2 - self.hinge_radius, 0, 0])

        body_children_nodes = new_make_block_nodes(
            scene,
            body_offset,
            self.size,
            make_parent=False,
            make_mesh=False,
            collision_shape_name=f"{door_body_node_name}_collision_shape",
        )
        body_mesh_node = get_scaled_mesh_node(
            scene, "res://entities/doors/door_body.tscn", self.size, body_offset, mesh_name="body_mesh"
        )
        body_children_nodes.append(body_mesh_node)
        handle_nodes = self._get_handle_nodes(scene)

        body_script = scene.add_ext_resource("res://entities/doors/door_body.gd", "Script")
        body_node = GDNode(
            door_body_node_name,
            type="RigidBody",
            properties={
                "transform": make_transform(position=body_position),
                "mass": 100.0,
                "script": body_script.reference,
                "is_auto_latching": self.latching_mechanics == LatchingMechanics.AUTO_LATCH,
                "entity_id": 0,
            },
        )
        for node in body_children_nodes:
            body_node.add_child(node)
        for node in handle_nodes:
            body_node.add_child(node)

        return [body_node]

    def _get_handle_nodes(self, scene: GodotScene) -> List[GDNode]:
        DOOR_HANDLE = "handle"

        door_width, door_height, door_thickness = self.size
        handle_width = door_width * self.handle_size_proportion
        handle_height = door_height * (self.handle_size_proportion / 2)
        handle_thickness = 0.15

        # We cap the handle height if there are locks that would get obstructed by it
        leeway = 0.05
        bar_locks = [lock for lock in self.locks if isinstance(lock, (RotatingBar, SlidingBar))]
        if len(bar_locks) > 0:
            max_y = np.inf
            min_y = -np.inf
            for lock in bar_locks:
                if not isinstance(lock, (RotatingBar, SlidingBar)):
                    continue
                if lock.position[1] > 0 and (lock_bottom := lock.position[1] - lock.size[1] / 2) < max_y:
                    max_y = lock_bottom
                elif lock.position[1] < 0 and (lock_top := lock.position[1] + lock.size[1] / 2) > min_y:
                    min_y = lock_top
            max_handle_height = min(abs(max_y), abs(min_y)) * 2 - leeway
            if handle_height > max_handle_height:
                handle_height = max_handle_height

        # We cap the handle width to ensure it doesn't prevent the door from opening inwards
        max_width = math.sqrt(
            (door_width - leeway) ** 2 - handle_thickness**2
        )  # diagonal of handle can't exceed door width
        handle_width = min(max_width, handle_width)
        handle_size = np.array([handle_width, handle_height, handle_thickness])

        handle_margin_skew = 25  # how many right-margins fit in the left margin
        handle_side_margin = (door_width - handle_width) / (1 + handle_margin_skew) * handle_margin_skew
        if self.hinge_side == HingeSide.LEFT:
            handle_x = handle_side_margin + handle_width / 2
        else:
            handle_x = -(handle_side_margin + handle_width / 2)

        handle_offsets = 1, -1
        handle_nodes = []
        for i, handle_offset in enumerate(handle_offsets):
            handle_position = np.array(
                [
                    handle_x,
                    0,
                    handle_offset * (handle_thickness / 2 + door_thickness / 2),
                ]
            )
            handle_name = f"{DOOR_HANDLE}_{i}"
            handle_nodes.extend(
                new_make_block_nodes(
                    scene,
                    handle_position,
                    handle_size,
                    make_parent=False,
                    make_mesh=False,
                    collision_shape_name=f"{handle_name}_collision_shape",
                )
            )
            handle_mesh_node = get_scaled_mesh_node(
                scene,
                "res://entities/doors/door_handle_loop.tscn",
                handle_size,
                handle_position,
                mesh_name=f"{handle_name}_mesh",
            )
            handle_nodes.append(handle_mesh_node)
        return handle_nodes

    def _get_hinge_nodes(self, scene: GodotScene, door_body_node_name: str) -> List[GDNode]:
        DOOR_HINGE = "hinge"

        door_width, door_height, door_thickness = self.size
        frame_width = door_width + 2 * self.hinge_radius
        hinge_height = door_height
        body_left_side = -frame_width / 2 + 2 * self.hinge_radius
        body_right_side = frame_width / 2 - 2 * self.hinge_radius

        if self.hinge_side == HingeSide.LEFT:
            hinge_x = -frame_width / 2 + self.hinge_radius
        else:
            hinge_x = frame_width / 2 - self.hinge_radius
        hinge_position = hinge_x, 0, 0
        hinge_material = scene.add_sub_resource(
            "SpatialMaterial", albedo_color=Color(0.223529, 0.196078, 0.141176, 1), roughness=0.2, metallic=0.5
        )
        hinge_mesh_node = GDNode(
            "mesh",
            "MeshInstance",
            properties={
                "transform": make_transform(),
                "mesh": scene.add_sub_resource(
                    "CylinderMesh",
                    material=hinge_material.reference,
                    top_radius=self.hinge_radius,
                    bottom_radius=self.hinge_radius,
                    height=hinge_height,
                ).reference,
            },
        )
        hinge_collision_mesh_node = GDNode(
            "collision_shape",
            "CollisionShape",
            properties={
                "shape": scene.add_sub_resource(
                    "CylinderShape", radius=self.hinge_radius, height=hinge_height
                ).reference,
                "disabled": True,
            },
        )
        hinge_node = GDNode(
            DOOR_HINGE,
            "StaticBody",
            properties={
                "transform": make_transform(position=hinge_position),
            },
        )
        hinge_node.add_child(hinge_mesh_node)
        hinge_node.add_child(hinge_collision_mesh_node)

        hinge_rotation_degrees = 90 if self.hinge_side == HingeSide.LEFT else -90
        hinge_joint_position = body_left_side if self.hinge_side == HingeSide.LEFT else body_right_side, 0, 0
        hinge_rotation = Rotation.from_euler("x", hinge_rotation_degrees, degrees=True).as_matrix().flatten()
        hinge_joint_node = GDNode(
            "hinge_joint",
            "HingeJoint",
            properties={
                "transform": make_transform(position=hinge_joint_position, rotation=hinge_rotation),
                "nodes/node_a": NodePath(f"../{DOOR_HINGE}"),
                "nodes/node_b": NodePath(f"../{door_body_node_name}"),
                "params/bias": 0.99,
                "angular_limit/enable": True,
                "angular_limit/lower": -self.max_outwards_angle,
                "angular_limit/upper": self.max_inwards_angle,
                "angular_limit/bias": 0.99,
                "angular_limit/softness": 0.01,
                "angular_limit/relaxation": 0.01,
            },
        )
        return [hinge_node, hinge_joint_node]



import filecmp
import gzip
import os
import shutil
import tarfile
from pathlib import Path
from typing import IO
from typing import cast

import numpy as np
from godot_parser import Node as GDNode
from loguru import logger

from avalon.common.utils import TEMP_DIR
from avalon.contrib.s3_utils import SimpleS3Client
from avalon.contrib.utils import create_temp_file_path
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.export import get_eval_agent_export_config
from avalon.datagen.world_creation.configs.export import get_oculus_export_config
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.indoor.builders import DefaultEntranceBuilder
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.task_generator import create_building_for_skill_scenario
from avalon.datagen.world_creation.indoor.task_generator import make_indoor_task_world
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.world_generator import GENERATION_FUNCTION_BY_TASK
from avalon.datagen.world_creation.worlds.compositional import get_building_task_generator_class
from avalon.datagen.world_creation.worlds.export import export_world


def export_building(building: Building, building_name: str, export_path: Path) -> None:
    scene = GodotScene()
    root_node = GDNode("root", "Spatial")
    with scene.use_tree() as tree:
        tree.root = root_node
        building.export(scene, root_node, building_name)
    scene.write(str(export_path.absolute()))


def reset_unix_epoch_mtime(tar_info: tarfile.TarInfo) -> tarfile.TarInfo:
    tar_info.uid = tar_info.gid = 0
    tar_info.uname = tar_info.gname = "user"
    tar_info.mtime = 1
    return tar_info


def make_tarball(root_path: Path, glob_pattern: str, archive_path: Path) -> None:
    last_pwd = os.getcwd()
    try:
        with cast(IO[bytes], gzip.GzipFile(archive_path, mode="wb", mtime=1)) as gzip_file:
            with tarfile.open("a.tar", "w", fileobj=gzip_file) as tarball:
                os.chdir(root_path)
                for entry in root_path.glob(glob_pattern):
                    entry_path = entry.relative_to(root_path)
                    tarball.add(entry_path, filter=reset_unix_epoch_mtime)
    finally:
        os.chdir(last_pwd)


def create_indoor_world(
    rand: np.random.Generator,
    difficulty: float,
    building_task: BuildingTask,
    radius: float,
    location: Point3DNP,
    export_path: Path,
    export_config: ExportConfig = get_oculus_export_config(),
) -> None:
    task_generator = get_building_task_generator_class(building_task)()
    building, entities, spawn_location, target_location = create_building_for_skill_scenario(
        rand,
        difficulty,
        task_generator,
        location,
        site_radius=radius,
        is_indoor_only=True,
    )
    with export_config.mutable_clone() as final_export_config:
        final_export_config.is_tiled = False
    export_config = final_export_config
    world = make_indoor_task_world(
        building, entities, difficulty, spawn_location, target_location, rand, final_export_config
    )
    export_world(export_path, rand, world)


def add_entrance_if_has_none(building: Building, rand: np.random.Generator) -> None:
    if not any([len(story.entrances) > 0 for story in building.stories]):
        entrance_builder = DefaultEntranceBuilder(top=False, bottom=True)
        entrance_builder.add_entrance(building.stories[0], rand)
        assert len(building.stories[0].entrances) == 1


def get_reference_data_path(reference_name: str, snapshot_commit: str) -> Path:
    qualified_reference_name = f"{reference_name}_reference"
    cached_reference_path = Path(TEMP_DIR) / f"{qualified_reference_name}__{snapshot_commit}"
    if not cached_reference_path.exists():
        s3_client = SimpleS3Client()
        archive_path = cached_reference_path.parent / f"{cached_reference_path.name}.tar.gz"
        logger.info(f"Reference {reference_name} not found locally, fetching from S3...")
        s3_client.download_to_file(f"{snapshot_commit}/{qualified_reference_name}.tar.gz", archive_path)
        cached_reference_path.mkdir()
        shutil.unpack_archive(archive_path, cached_reference_path)
    return cached_reference_path


def compare_files(generated_file_path: Path, reference_file_path: Path) -> None:
    # Comparing the file contents directly generates a diff too big for Pycharm to view directly; instead we
    # store a copy of the files and output the paths, which you can throw into the diff tool yourself
    if not filecmp.cmp(generated_file_path, reference_file_path):
        with create_temp_file_path(cleanup=False) as generated_file_copy_path, create_temp_file_path(
            cleanup=False
        ) as reference_file_copy_path:
            shutil.copyfile(generated_file_path, generated_file_copy_path)
            shutil.copyfile(reference_file_path, reference_file_copy_path)
            assert False, (
                f"Generated and reference files ({generated_file_path.name}) don't match\n"
                f"Generated: {generated_file_copy_path}\nReference: {reference_file_copy_path}"
            )


def create_world(
    task: AvalonTask,
    difficulty: float,
    seed: int,
    export_path: Path,
    export_config: ExportConfig = get_eval_agent_export_config(),
) -> Path:
    # todo: used commonly - generalize, push up, replace other calls?
    rand = np.random.default_rng(seed)
    generation_function = GENERATION_FUNCTION_BY_TASK[task]
    try:
        generation_function(rand, difficulty, export_path, export_config)
    except ImpossibleWorldError:
        logger.error(
            f"Got ImpossibleWorldError while generating world for {task=}, {difficulty=}, {seed=}, path will be empty"
        )
    except Exception as e:
        # logger.error(f"Got '{type(e)}: {e.args}' while generating world for {task=}, {difficulty=}, {seed=}")
        raise
    return export_path


GENERIC_PATH_PLACEHOLDER = "<replaced for reproducibility>"


def make_world_file_generic(path: Path) -> None:
    with open(path, "r+") as file:
        data = file.read()
        file.seek(0)
        file.truncate()
        file.write(data.replace(str(path.parent), GENERIC_PATH_PLACEHOLDER))



import json
import multiprocessing as mp
import shutil
import signal
import time
import uuid
import warnings
from multiprocessing import Lock
from multiprocessing import Pool
from pathlib import Path
from random import Random
from typing import Any
from typing import ClassVar
from typing import Dict
from typing import Final
from typing import Generic
from typing import Hashable
from typing import List
from typing import Optional
from typing import Protocol
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

import attr
import numpy as np
from godot_parser import Node as GDNode
from typing_extensions import TypeGuard

from avalon.common.errors import SwitchError
from avalon.common.log_utils import configure_parent_logging
from avalon.common.log_utils import logger
from avalon.contrib.serialization import Serializable
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.export import get_agent_export_config
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import TASKS_BY_TASK_GROUP
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.constants import AvalonTaskGroup
from avalon.datagen.world_creation.constants import get_all_tasks_for_task_groups
from avalon.datagen.world_creation.entities.spawn_point import PLAYER_SPAWN_POINT
from avalon.datagen.world_creation.tasks.avoid import generate_avoid_task
from avalon.datagen.world_creation.tasks.bridge import generate_bridge_task
from avalon.datagen.world_creation.tasks.carry import generate_carry_task
from avalon.datagen.world_creation.tasks.climb import generate_climb_task
from avalon.datagen.world_creation.tasks.descend import generate_descend_task
from avalon.datagen.world_creation.tasks.eat import generate_eat_task
from avalon.datagen.world_creation.tasks.explore import generate_explore_task
from avalon.datagen.world_creation.tasks.fight import generate_fight_task
from avalon.datagen.world_creation.tasks.find import generate_find_task
from avalon.datagen.world_creation.tasks.gather import generate_gather_task
from avalon.datagen.world_creation.tasks.hunt import generate_hunt_task
from avalon.datagen.world_creation.tasks.jump import generate_jump_task
from avalon.datagen.world_creation.tasks.move import generate_move_task
from avalon.datagen.world_creation.tasks.navigate import generate_navigate_task
from avalon.datagen.world_creation.tasks.open import generate_open_task
from avalon.datagen.world_creation.tasks.push import generate_push_task
from avalon.datagen.world_creation.tasks.scramble import generate_scramble_task
from avalon.datagen.world_creation.tasks.stack import generate_stack_task
from avalon.datagen.world_creation.tasks.survive import generate_survive_task
from avalon.datagen.world_creation.tasks.throw import generate_throw_task
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.worlds.difficulty import select_categorical_difficulty

DEFAULT_WORLD_ID: Final = 0


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class GenerateWorldParams(Serializable):
    output: str

    @property
    def output_path(self) -> Path:
        return Path(self.output)

    @property
    def main_scene_path(self) -> str:
        return f"{self.output}/main.tscn"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class GeneratedWorldParams(GenerateWorldParams):
    @classmethod
    def from_input(cls, input_params: GenerateWorldParams) -> "GeneratedWorldParams":
        return cls(**attr.asdict(input_params))


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class GenerateAvalonWorldParams(GenerateWorldParams):
    task: AvalonTask
    difficulty: float
    seed: int
    index: int
    output: str
    # this parameter is a trade-off between how quickly you will hear about a task generator that is
    # totally busted (lower values) vs the probability that your process crashed because you just got unlucky
    # (higher values). Can set to inf if you're sure there are no bugs and you definitely want it to run until it finds a level...
    num_retries: int = 5


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class GeneratedAvalonWorldParams(GenerateAvalonWorldParams, GeneratedWorldParams):
    SERIALIZED_FILE_NAME: ClassVar = "params.json"

    @staticmethod
    def from_world_path(existing_world_path: Path) -> "GeneratedAvalonWorldParams":
        with open(existing_world_path / GeneratedAvalonWorldParams.SERIALIZED_FILE_NAME) as file:
            return GeneratedAvalonWorldParams.from_dict(json.load(file))

    def save_to_output_path(self) -> None:
        serialized_path = self.output_path / self.SERIALIZED_FILE_NAME
        with open(serialized_path, "w") as file:
            json.dump(self.to_dict(), file)


GenerateWorldParamsType = TypeVar("GenerateWorldParamsType", bound=GenerateWorldParams)
GeneratedWorldParamsType = TypeVar("GeneratedWorldParamsType", bound=GeneratedWorldParams)


class WorldGenerator(Generic[GeneratedWorldParamsType]):
    def __init__(self, base_path: Path, seed: int) -> None:
        self.output_path = base_path / str(uuid.uuid4())
        self.output_path.mkdir(parents=True, exist_ok=False)
        self.set_seed(seed)

    def generate_batch(self, start_world_id: Optional[int], batch_size: int = 1) -> List[GeneratedWorldParamsType]:
        """
        :returns: a (batch_size)-long list of folders (one folder per scene).
        The folder can be assumed to contain a "main.tscn" file
        """
        raise NotImplementedError()

    def set_seed(self, seed: int) -> None:
        self._seed = seed

    def close(self) -> None:
        if self.output_path.exists():
            shutil.rmtree(self.output_path)


class EmptyLevelGenerator(WorldGenerator[GeneratedWorldParams]):
    def __init__(self, base_path: Path, seed: int) -> None:
        super().__init__(base_path, seed)
        scene = GodotScene()
        with scene.use_tree() as tree:
            tree.root = GDNode("Avalon", type="Spatial")
            tree.root.add_child(GDNode(PLAYER_SPAWN_POINT, type="Spatial"))
            env = scene.add_sub_resource("Environment")
            env_node = GDNode(
                "WorldEnvironment",
                type="WorldEnvironment",
                properties=dict(environment=env.reference),
            )
            tree.root.add_child(env_node)
        scene.write(str(self.output_path.absolute() / "main.tscn"))

    def generate_batch(self, start_world_id: Optional[int], batch_size: int = 1) -> List[GeneratedWorldParams]:
        assert batch_size == 1
        return [GeneratedWorldParams(str(self.output_path))]


class AvalonWorldGenerator(WorldGenerator[GeneratedAvalonWorldParams]):
    def sample_tasks(self, n_tasks: int) -> List[Hashable]:
        """Samples task of the meta-environment

        Args:
            n_tasks (int) : number of different meta-tasks needed

        Returns:
            tasks (list) : an (n_tasks) length list of tasks
        """
        raise NotImplementedError()

    def set_task(self, task: Hashable) -> None:
        """Sets the specified task to the current environment

        Args:
            task: task of the meta-learning environment
        """
        raise NotImplementedError()

    def get_task(self) -> Hashable:
        """Gets the task that the agent is performing in the current environment

        Returns:
            task: task of the meta-learning environment
        """
        raise NotImplementedError()

    def load_already_generated_params(self, world_index: int) -> Optional[GeneratedAvalonWorldParams]:
        world_path = Path(self.output_path) / str(world_index)
        if not world_path.exists():
            return None
        return GeneratedAvalonWorldParams.from_world_path(world_path)


def is_fixed_world_generator(
    generator: AvalonWorldGenerator,
) -> TypeGuard[Union["FixedWorldGenerator", "FixedWorldLoader"]]:
    return isinstance(generator, (FixedWorldGenerator, FixedWorldLoader))


class SingleTaskWorldGenerator(AvalonWorldGenerator):
    def __init__(self, base_path: Path, seed: int, difficulty: float, task: AvalonTask) -> None:
        super().__init__(base_path, seed)
        self.difficulty = difficulty
        self.task = task

    def generate_batch(self, start_world_id: Optional[int], batch_size: int = 1) -> List[GeneratedAvalonWorldParams]:
        """
        :returns: a (batch_size)-long list of folders (one folder per scene).
        The folder can be assumed to contained a "main.tscn" file
        """
        raise NotImplementedError()

    def sample_tasks(self, n_tasks: int) -> List[Hashable]:
        return [0] * n_tasks

    def set_task(self, task: Hashable) -> None:
        pass

    def get_task(self) -> int:
        return 0


class InsufficientBufferSize(Exception):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class WorldGeneratorTask:
    task: AvalonTask
    difficulty: float


def rand_task_with_difficulty(
    task_groups: Tuple[AvalonTaskGroup, ...], difficulty: float, rand: np.random.Generator
) -> AvalonTask:
    task_group, _difficulty = select_categorical_difficulty(task_groups, difficulty, rand)
    return rand.choice(TASKS_BY_TASK_GROUP[task_group])  # type: ignore


class BlockingWorldGenerator(AvalonWorldGenerator):
    def __init__(
        self,
        base_path: Path,
        seed: int,
        start_difficulty: float,
        task_groups: Tuple[AvalonTaskGroup, ...],
    ) -> None:
        super().__init__(base_path, seed)
        self.difficulties = {x: start_difficulty for x in get_all_tasks_for_task_groups(task_groups)}
        self.task_groups = task_groups
        self.meta_difficulty = start_difficulty

    def _generate_world(self, i: int) -> GeneratedAvalonWorldParams:
        rand = np.random.default_rng([self._seed, i])
        task = rand_task_with_difficulty(self.task_groups, self.meta_difficulty, rand)

        return generate_world(
            GenerateAvalonWorldParams(
                task=task,
                difficulty=self.difficulties[task],
                seed=self._seed,
                index=i,
                output=str(self.output_path / str(i)),
            )
        )

    def generate_batch(self, start_world_id: Optional[int], batch_size: int = 1) -> List[GeneratedAvalonWorldParams]:
        if start_world_id is None:
            start_world_id = DEFAULT_WORLD_ID
        return [self._generate_world(i) for i in range(start_world_id, start_world_id + batch_size)]

    def set_task(self, task: Hashable) -> None:
        assert isinstance(task, WorldGeneratorTask)
        self.task = task.task
        self.difficulties[task.task] = task.difficulty

    def set_task_difficulty(self, task: AvalonTask, difficulty: float) -> None:
        self.difficulties[task] = difficulty

    def set_meta_difficulty(self, difficulty: float) -> None:
        self.meta_difficulty = difficulty

    def get_task(self) -> WorldGeneratorTask:
        assert self.task is not None
        return WorldGeneratorTask(task=self.task, difficulty=self.difficulties[self.task])

    def set_seed(self, seed: int) -> None:
        super().set_seed(seed)


def get_world_params_for_task_groups(
    task_groups: Tuple[AvalonTaskGroup, ...],
    difficulties: Tuple[float, ...],
    output_path: Path,
    seed: int,
) -> List[GenerateAvalonWorldParams]:
    worlds: List[GenerateAvalonWorldParams] = []
    index = 0
    for task in get_all_tasks_for_task_groups(task_groups):
        for difficulty in difficulties:
            worlds.append(
                GenerateAvalonWorldParams(
                    task=task,
                    difficulty=difficulty,
                    seed=seed,
                    index=index,
                    output=str(output_path / str(index)),
                )
            )
            index += 1

    return worlds


def generate_fixed_worlds(
    world_params: List[GenerateAvalonWorldParams],
    num_processes: int = 4,
) -> Dict[int, GeneratedAvalonWorldParams]:
    worlds: Dict[int, GeneratedAvalonWorldParams] = {}
    index = 0

    ctx = mp.get_context("spawn")

    def on_done(result: GeneratedAvalonWorldParams) -> None:
        # logger.info(f"Finished generating {result}")
        worlds[result.index] = result

    def on_error(error: BaseException):
        logger.error(f"World generation failed! {error}")
        raise error

    with ctx.Pool(
        processes=min(len(world_params), num_processes), initializer=configure_parent_logging
    ) as worker_pool:
        for params in world_params:
            worker_pool.apply_async(
                generate_world,
                kwds={"params": params},
                callback=on_done,
                error_callback=on_error,
            )
        worker_pool.close()
        while ctx.active_children():
            # For some reason, when using the PyCharm debugger, it hangs on `join()`, so I added this block.
            # Note: this might break if this process has other children? Not the case currently.
            if len(worlds) == len(world_params):
                worker_pool.terminate()
            time.sleep(1)
        worker_pool.join()

    return worlds


def _copy_world(world: GeneratedAvalonWorldParams, filename_suffix: str = "working") -> GeneratedAvalonWorldParams:
    original_path = world.output_path
    dest_path = original_path.with_stem(f"{original_path.stem}_{filename_suffix}")
    shutil.copytree(str(original_path), str(dest_path), dirs_exist_ok=True)
    return attr.evolve(world, output=str(dest_path))


class FixedWorldLoader(AvalonWorldGenerator):
    """Load worlds from the given path instead of generating them."""

    def __init__(
        self,
        base_path: Path,
        generated_worlds_path: Path,
        num_generators: int = 1,
        generator_index: int = 0,
    ) -> None:
        unused_placeholder_seed = -1
        super().__init__(base_path, seed=unused_placeholder_seed)
        self._init_by_loading(generated_worlds_path, num_generators, generator_index)

    def _init_by_loading(self, load_path: Path, num_generators: int, generator_index: int):
        world_paths = sorted(path for path in load_path.iterdir() if not str(path).startswith("practice"))
        selected_list_indices_and_paths = [
            (i, path) for i, path in enumerate(world_paths) if i % num_generators == generator_index
        ]

        assert len(selected_list_indices_and_paths) > 0
        self.worlds: Dict[int, GeneratedAvalonWorldParams] = {}
        for list_index, world_path in selected_list_indices_and_paths:
            new_world_path = self.output_path / world_path.name
            shutil.copytree(world_path, new_world_path)
            world_name_parts = new_world_path.name.split("__")
            task = AvalonTask(world_name_parts[0].upper())
            world_index_and_seed = int(world_name_parts[1])
            if world_index_and_seed in self.worlds:
                raise Exception("Cannot have two worlds with the same index!")
            difficulty = float(world_name_parts[2].replace("_", "."))

            # TODO should we use GeneratedAvalonWorldParams.from_world_path?
            # Don't liket hat the seed is a lie
            world = GeneratedAvalonWorldParams(
                task=task,
                difficulty=difficulty,
                seed=world_index_and_seed,
                index=world_index_and_seed,
                output=str(new_world_path),
            )
            self.worlds[world.index] = world

    def generate_batch(self, start_world_id: Optional[int], batch_size: int = 1) -> List[GeneratedAvalonWorldParams]:
        assert batch_size == 1, "Not supported for FixedWorldLoader"
        if start_world_id is None:
            start_world_id = DEFAULT_WORLD_ID
        return [_copy_world(self.worlds[start_world_id])]


class FixedWorldGenerator(AvalonWorldGenerator):
    """Generates num_unique_worlds and copies them to a working dir each time they are requested."""

    def __init__(
        self,
        base_path: Path,
        seed: int,
        difficulties: Tuple[float, ...],
        task_groups: Tuple[AvalonTaskGroup, ...],
        num_generators: int = 1,
        generator_index: int = 0,
    ) -> None:
        super().__init__(base_path, seed)
        self._init_by_generating(difficulties, task_groups, num_generators, generator_index)

    def _init_by_generating(
        self,
        difficulties: Tuple[float, ...],
        task_groups: Tuple[AvalonTaskGroup, ...],
        num_generators: int,
        generator_index: int,
    ) -> None:
        world_params = get_world_params_for_task_groups(
            task_groups,
            difficulties,
            self.output_path,
            self._seed,
        )
        # each generator will be responsible for worlds where index % num_generators == generator_index
        # TODO: should (deterministically) shuffle these to ensure an unbiased split (in length of episodes)
        world_params = [x for x in world_params if x.index % num_generators == generator_index]
        assert len(world_params) > 0
        self.worlds = generate_fixed_worlds(world_params)

    def generate_batch(self, start_world_id: Optional[int], batch_size: int = 1) -> List[GeneratedAvalonWorldParams]:
        assert batch_size == 1, "Not supported for fixed world gen"
        if start_world_id is None:
            start_world_id = DEFAULT_WORLD_ID
        return [_copy_world(self.worlds[start_world_id])]


def disable_sigint() -> None:
    # Disable sigint handler so we can handle cleanup neatly ourselves.
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    configure_parent_logging()


# TODO: this introduces some non-determinism, fix it by anchoring to ids of levels generated
class LocalProcessWorldGenerator(AvalonWorldGenerator):
    def __init__(
        self,
        base_path: Path,
        seed: int,
        task_groups: Tuple[AvalonTaskGroup, ...],
        min_difficulty: float = 0,
        start_difficulty: float = 0,
        num_workers: int = 2,
        buffer_size: int = 20,
        is_task_curriculum_used: bool = True,
    ) -> None:
        super().__init__(base_path, seed)
        self.task_groups = task_groups
        self.min_difficulty = min_difficulty
        self.difficulties = {x: start_difficulty for x in get_all_tasks_for_task_groups(task_groups)}
        self.meta_difficulty = start_difficulty
        self.is_task_curriculum_used = is_task_curriculum_used
        self.rand = np.random.default_rng([self._seed])
        self.buffer: List[GeneratedAvalonWorldParams] = []
        self.lock = Lock()
        self.worker_pool = Pool(processes=num_workers, initializer=disable_sigint)
        self.offset = 0  # roughly, stores the index of the world to be generated next.
        self._request_batch(buffer_size)

    def generate_batch(self, start_world_id: Optional[int], batch_size: int = 1) -> List[GeneratedAvalonWorldParams]:
        assert start_world_id is None, "Not supported for local process world gen"
        self._request_batch(batch_size)
        while True:
            try:
                batch = self._pop_batch(batch_size)
                return batch
            except InsufficientBufferSize:
                time.sleep(0.1)

    def close(self) -> None:
        if self.worker_pool is not None:
            self._destroy_pool()
        super().close()

    def _destroy_pool(self) -> None:
        self.worker_pool.close()
        # These are pretty quick, just let them complete.
        # self.worker_pool.terminate()
        self.worker_pool.join()
        self.worker_pool = None  # type: ignore

    def _request_batch(self, batch_size: int) -> None:
        with self.lock:
            sampled_tasks = [
                rand_task_with_difficulty(self.task_groups, self.meta_difficulty, self.rand) for _ in range(batch_size)
            ]
            for idx in range(batch_size):
                i = idx + self.offset
                task = sampled_tasks[idx]
                output_dir = self.output_path / str(i)
                assert (
                    not output_dir.exists()
                ), f"{output_dir} already exists! Seed {self._seed} duplicated or path not cleared?"
                if self.is_task_curriculum_used:
                    difficulty = self.rand.uniform(low=self.min_difficulty, high=self.difficulties[task])
                else:
                    difficulty = self.rand.uniform()
                self.worker_pool.apply_async(
                    generate_world,
                    args=(
                        GenerateAvalonWorldParams(
                            task=task,
                            difficulty=difficulty,
                            seed=self._seed,
                            index=i,
                            output=str(output_dir),
                        ),
                    ),
                    callback=self._on_world_generation_done,
                    error_callback=self._on_world_generation_error,
                )
            self.offset += batch_size

    def _pop_batch(self, batch_size: int) -> List[GeneratedAvalonWorldParams]:
        with self.lock:
            if len(self.buffer) >= batch_size:
                result = self.buffer[:batch_size]
                self.buffer = self.buffer[batch_size:]
                return result
            raise InsufficientBufferSize()

    def _on_world_generation_done(self, result: GeneratedAvalonWorldParams) -> None:
        with self.lock:
            self.buffer.append(result)

    def _on_world_generation_error(self, error: BaseException) -> None:
        logger.error("Caught error in _on_world_generation_error. This shouldn't happen!!!:")
        logger.error(error)

    def set_task_difficulty(self, task: AvalonTask, difficulty: float) -> None:
        with self.lock:
            self.difficulties[task] = max(self.min_difficulty, difficulty)

    def set_meta_difficulty(self, difficulty: float) -> None:
        with self.lock:
            self.meta_difficulty = difficulty

    def set_seed(self, seed: int) -> None:
        super().set_seed(seed)
        # TODO: can this be a np.random.Generator as well?
        self.rand = Random(f"seed:{self._seed}")  # type: ignore


class TaskGenerationFunction(Protocol):
    def __call__(
        self,
        rand: np.random.Generator,
        difficulty: float,
        output_path: Path,
        export_config: ExportConfig,
        task_config: Any = TaskConfig(),
    ) -> None:
        ...


GENERATION_FUNCTION_BY_TASK: Final[Dict[AvalonTask, TaskGenerationFunction]] = {
    AvalonTask.EAT: generate_eat_task,
    AvalonTask.MOVE: generate_move_task,
    AvalonTask.JUMP: generate_jump_task,
    AvalonTask.CLIMB: generate_climb_task,
    AvalonTask.DESCEND: generate_descend_task,
    AvalonTask.STACK: generate_stack_task,
    AvalonTask.BRIDGE: generate_bridge_task,
    AvalonTask.PUSH: generate_push_task,
    AvalonTask.THROW: generate_throw_task,
    AvalonTask.HUNT: generate_hunt_task,
    AvalonTask.FIGHT: generate_fight_task,
    AvalonTask.AVOID: generate_avoid_task,
    AvalonTask.NAVIGATE: generate_navigate_task,
    AvalonTask.FIND: generate_find_task,
    AvalonTask.EXPLORE: generate_explore_task,
    AvalonTask.SURVIVE: generate_survive_task,
    AvalonTask.OPEN: generate_open_task,
    AvalonTask.CARRY: generate_carry_task,
    AvalonTask.GATHER: generate_gather_task,
    AvalonTask.SCRAMBLE: generate_scramble_task,
}

MAX_NOISE_SCALE = 0.3


@logger.catch(reraise=True)
def generate_world(params: GenerateAvalonWorldParams) -> GeneratedAvalonWorldParams:
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    output_path = Path(params.output)
    output_path.mkdir(parents=True)
    rand = np.random.default_rng([params.seed, params.index])
    export_config = get_agent_export_config()
    generation_function = GENERATION_FUNCTION_BY_TASK[params.task]
    for i in range(params.num_retries):
        try:
            generation_function(rand, params.difficulty, output_path, export_config)
            generated_params = cast(GeneratedAvalonWorldParams, GeneratedAvalonWorldParams.from_input(params))
            generated_params.save_to_output_path()
            return generated_params
        except ImpossibleWorldError as e:
            if i == params.num_retries - 1:
                logger.error("Ran out of retries to generate a good world!")
                logger.error(f"Params were: {params}")
                raise
            else:
                logger.debug(f"Impossible world was generated, this was try {i}... (reason: {e})")
                # clear the output
                shutil.rmtree(output_path)
                output_path.mkdir(parents=True, exist_ok=True)
        except Exception as e:
            if i == params.num_retries - 1:
                logger.error("Ran out of retries to generate a good world!")
                logger.error(f"Params were: {params}")
                raise
            else:
                logger.error(f"Unspecified error in world generation, this was try {i}... (reason: {e})")
                # clear the output
                shutil.rmtree(output_path)
                output_path.mkdir(parents=True, exist_ok=True)
    raise SwitchError("This should never happen")


if __name__ == "__main__":
    generate_world(
        GenerateAvalonWorldParams(
            AvalonTask.MOVE,
            1,
            42,
            0,
            str(Path("./standalone/avalon/datagen/godot")),
        )
    )



from typing import Tuple

import numpy as np

from avalon.datagen.world_creation.debug_plots import plot_value_grid
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.worlds.obstacles.harmonics import create_harmonics
from avalon.datagen.world_creation.worlds.obstacles.ring_obstacle import HeightObstacle
from avalon.datagen.world_creation.worlds.obstacles.ring_obstacle import RingObstacle


def create_obstacle_masks(
    rand: np.random.Generator,
    ring: RingObstacle,
    config: HeightObstacle,
    island_mask: MapBoolNP,
    # whether debugging is enabled
    is_debug_graph_printing_enabled: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    The high level purpose of this function is to create obstacles (chasms, cliffs, and ridges).

    Using a chasm as a concrete example, we want to be able to finely control the exact distance
    that an agent would need to jump in order to cross, in addition to what fraction of the chasm
    is crossable at that width.

    Of course, we could simply draw a line or a circle for the chasm which is exactly the right size,
    but this looks extremely unnatural.

    The purpose of this function is to make it easy to create more natural looking chasms.

    At a high level, it works as follows:
    - Use spherical harmonic noise to make a wiggly circle that represents the bottom middle of the chasm
    - on either side, add purely additive spherical harmonic noise in order to vary the sizes of the chasm
        With respect to theta, this noise is varied smoothly such that within traversal_width / 2.0 of the
        traversal point, there is no noise, and at (traversal_width / 2.0) *  traversal_noise_interpolation_multiple
        the noise is the maximum value (which is defined by the safety_radius of the inner and outer rings)
    - linearly interpolate the transitions between the bottom and top of the chasm over edge_distance meters (roughly)
    """

    if config.is_inside_ring:
        assert config.traversal_theta_offset == 0.0, "Not allowed to vary theta inside the ring"

    # figure out where the traversal point is in polar coordinates
    traversal_offset = ring.config.traversal_point - ring.config.center_point
    traversal_theta = np.arctan2(traversal_offset[1], traversal_offset[0]) + config.traversal_theta_offset

    # figure out the interpolation bounds for our traversal
    # they are represented in "normalized theta delta" space, eg, 0.5 means "90 degrees away from the traversal
    # angle"
    traversal_distance = np.sqrt(np.sum((ring.config.center_point - ring.config.traversal_point) ** 2))
    if config.is_inside_ring:
        distance_delta = -config.traversal_length
    else:
        distance_delta = config.traversal_length + ring.config.chasm_bottom_size
    traversal_width_theta = np.arctan2(config.traversal_width / 2.0, traversal_distance + distance_delta)
    theta_width_interp_start = traversal_width_theta / np.pi
    theta_width_interp_end = min(config.traversal_noise_interpolation_multiple * theta_width_interp_start, 1.0)
    theta_interp_width = theta_width_interp_end - theta_width_interp_start

    # create the normalized theta space
    normalized_theta_delta = ring.theta.copy()
    normalized_theta_delta -= traversal_theta
    normalized_theta_delta[normalized_theta_delta < -np.pi] += 2 * np.pi
    normalized_theta_delta[normalized_theta_delta > np.pi] -= 2 * np.pi
    normalized_theta_delta[normalized_theta_delta > 0] *= -1.0
    normalized_theta_delta = (normalized_theta_delta / np.pi) + 1

    # create the mask (0 - 1) that can perform the actual interpolation
    theta_mask = normalized_theta_delta.copy()
    theta_mask[normalized_theta_delta >= theta_width_interp_end] = 1.0
    theta_mask[normalized_theta_delta < theta_width_interp_end] = 0.0
    theta_region = np.logical_and(
        normalized_theta_delta >= theta_width_interp_start, normalized_theta_delta < theta_width_interp_end
    )
    theta_mask[theta_region] = (normalized_theta_delta[theta_region] - theta_width_interp_start) / theta_interp_width

    # create the output obstacle mask (which we will update below)
    # 0.0 will represent the region that we start in, and 1.0 the region of the obstacle
    obstacle_mask = ring.z.copy()

    # TODO: actually tie this constant to the other
    # this is intimately related to how many grid points per meter there are
    MARGIN = 0.25

    if config.is_inside_ring:
        # figure out the maximum z that could be added without exceeding the safety radius
        safety_radius = ring.config.inner_safety_radius
        outline = np.bitwise_and(safety_radius + MARGIN > ring.r, ring.r > safety_radius - MARGIN)
        if is_debug_graph_printing_enabled:
            plot_value_grid(outline, title="Inner safety radius")
        z_at_safety_radius = ring.z[outline].max()
        max_z_to_add = (ring.mid_z - config.traversal_length) - z_at_safety_radius
        if max_z_to_add <= 0.0:
            max_z_to_add = 0.01

        # calculate the noised up inner edge and bottom cutoffs
        harmonics = (
            create_harmonics(rand, ring.theta, config.edge_config.to_harmonics(traversal_distance), is_normalized=True)
            * max_z_to_add
        )
        inner_theta_dependent_noise = harmonics * theta_mask
        inner_edge_z = (ring.mid_z - config.traversal_length) - inner_theta_dependent_noise
        inner_bottom_z = ring.mid_z - inner_theta_dependent_noise

        # everything less than the inner edge is the inner region, mask should be zero
        inner_ring_inner_region = ring.z < inner_edge_z
        obstacle_mask[inner_ring_inner_region] = 0.0

        if config.traversal_length > 0.01:
            # interpolate smoothly between inner_edge_z and inner_bottom_z
            interpolating_region = np.logical_and(ring.z >= inner_edge_z, ring.z < inner_bottom_z)
            obstacle_mask[interpolating_region] = ((ring.z - inner_edge_z) / (inner_bottom_z - inner_edge_z))[
                interpolating_region
            ]
            if is_debug_graph_printing_enabled:
                plot_value_grid(interpolating_region, title="Inner ring interpolation region")

        # if there is no outer ring, we're basically done--just clamp the rest of the values.
        # if outer_ring is None:
        obstacle_mask = np.clip(obstacle_mask, 0, 1)

        inner_region = inner_ring_inner_region
        outer_region = ring.z > inner_bottom_z
    else:
        # figure out the maximum z that could be added without exceeding the safety radius
        safety_radius = ring.config.outer_safety_radius
        outline = np.bitwise_and(safety_radius + MARGIN > ring.r, ring.r > safety_radius - MARGIN)
        if is_debug_graph_printing_enabled:
            plot_value_grid(outline, title="Outer safety radius")
        z_at_safety_radius = ring.z[outline].min()
        max_z_to_add = z_at_safety_radius - (ring.mid_z + config.traversal_length)
        # logger.debug(max_z_to_add)
        if max_z_to_add <= 0.0:
            max_z_to_add = 0.01

        # calculate the noised up outer edge and bottom cutoffs
        harmonics = (
            create_harmonics(rand, ring.theta, config.edge_config.to_harmonics(traversal_distance), is_normalized=True)
            * max_z_to_add
        )
        outer_theta_dependent_noise = harmonics * theta_mask
        outer_bottom_z = ring.mid_z + outer_theta_dependent_noise
        outer_edge_z = (ring.mid_z + config.traversal_length) + outer_theta_dependent_noise

        # everything greater than the outer edge is the outer region, mask should be zero
        outer_ring_outer_region = ring.z > outer_edge_z
        obstacle_mask[outer_ring_outer_region] = 0.0

        if config.traversal_length > 0.01:
            # interpolate smoothly between outer_bottom_z and outer_edge_z
            interpolating_region = np.logical_and(ring.z >= outer_bottom_z, ring.z < outer_edge_z)
            obstacle_mask[interpolating_region] = ((outer_edge_z - ring.z) / (outer_edge_z - outer_bottom_z))[
                interpolating_region
            ]
            if is_debug_graph_printing_enabled:
                plot_value_grid(interpolating_region, title="Outer ring interpolation region")

        # if there is no inner ring, we're basically done
        # if inner_ring is None:
        # we just set anything in the inner region to one as well (since we're going to invert below)
        obstacle_mask[ring.z < ring.mid_z] = 1.0
        # invert and clip
        obstacle_mask = 1.0 - np.clip(obstacle_mask, 0, 1)

        inner_region = ring.z < outer_bottom_z
        outer_region = outer_ring_outer_region

    # # if both the inner and outer rings are define, make sure we set all of the values in the bottom of the chasm
    # # to 1.0
    # if inner_ring and outer_ring:
    #     chasm_bottom = np.logical_and(ring.z >= inner_bottom_z, ring.z < outer_bottom_z)
    #     obstacle_mask[chasm_bottom] = 1.0
    #     if is_debug_graph_printing_enabled:
    #         plot_value_grid(chasm_bottom, title="Chasm bottom")
    #     inner_region = inner_ring_inner_region
    #     outer_region = outer_ring_outer_region
    # elif inner_ring:
    #     inner_region = inner_ring_inner_region
    #     outer_region = ring.z > inner_bottom_z
    # else:
    #     assert outer_ring
    #     inner_region = ring.z < outer_bottom_z
    #     outer_region = outer_ring_outer_region

    # print the result if debugging is enabled
    if is_debug_graph_printing_enabled:
        plot_value_grid(inner_region, title="Inner region (full)")
        plot_value_grid(outer_region, title="Outer region (full)")
        plot_value_grid(obstacle_mask, title="Final mask (full)")

    not_island = np.logical_not(island_mask)
    obstacle_mask[not_island] = 0.0
    inner_region[not_island] = 0
    outer_region[not_island] = 0

    if is_debug_graph_printing_enabled:
        plot_value_grid(inner_region, title="Inner region (minus island)")
        plot_value_grid(outer_region, title="Outer region (minus island)")
        plot_value_grid(obstacle_mask, title="Final mask (minus island)")

    return obstacle_mask, inner_region, outer_region



"""
Script to evaluate participant's submission and report to EvalAI.
Modified from: https://github.com/Cloud-CV/EvalAI-Starters/blob/master/remote_challenge_evaluation/evaluation_script_starter.py
"""
import base64
import json
import os
import re
import subprocess
import time
import urllib.request
from typing import Any
from typing import Dict

import boto3
import docker
from loguru import logger

from avalon.contest.test_runner.evalai_interface import EvalAI_Interface

# IDs (generated by EvalAI) to represent different challenge phases
PHASE_MINIVAL = 3747
PHASE_PUBLIC_TEST = 3748
PHASE_PRIVATE_TEST = 3773


def login_docker_client_to_aws_ecr() -> None:
    """
    Login to ECR, with a workaround for bug in docker SDK.
    See: https://stackoverflow.com/questions/53759374/docker-login-to-ecr-via-python-docker-sdk
    """
    ecr_client = boto3.client(
        "ecr",
        aws_access_key_id=os.getenv("EVALAI_AWS_ACCESS_KEY_ID"),
        aws_secret_access_key=os.getenv("EVALAI_AWS_SECRET_ACCESS_KEY"),
        region_name=os.getenv("EVALAI_REGION_NAME"),
    )

    token = ecr_client.get_authorization_token()
    username, password = base64.b64decode(token["authorizationData"][0]["authorizationToken"]).decode().split(":")
    registry = token["authorizationData"][0]["proxyEndpoint"]

    command = 'docker login -u "%s" -p "%s" "%s"' % (username, password, registry)

    p = subprocess.Popen([command], stdout=subprocess.PIPE, shell=True, bufsize=1)
    for line in iter(p.stdout.readline, b""):  # type: ignore
        logger.info(line)
    p.communicate()  # close p.stdout, wait for the subprocess to exit


def handle_submission(evalai: EvalAI_Interface, submission: Dict[str, Any]) -> None:
    input_file_url = submission["input_file"]

    with urllib.request.urlopen(input_file_url) as url:
        submitted_image_uri = json.load(url)["submitted_image_uri"]

    repository, tag = submitted_image_uri.split(":")
    repository = repository.replace("https://", "")

    logger.info("Retrieving image:", submitted_image_uri)
    logger.info("Repository:", repository)
    logger.info("Tag:", tag)

    login_docker_client_to_aws_ecr()

    docker_client = docker.from_env()
    submitted_image = docker_client.images.pull(repository=repository, tag=tag)

    # Update EvalAI right after sending the submission into "RUNNING" state
    evalai.update_submission_status({"submission": submission["id"], "submission_status": "RUNNING"})

    try:
        logger.info(f"Running submission: {submitted_image_uri}")

        if submission["challenge_phase"] == PHASE_MINIVAL:
            container_environment = {"FIXED_WORLDS_PATH": "/tmp/avalon_worlds/minival/"}
        elif submission["challenge_phase"] == PHASE_PUBLIC_TEST:
            container_environment = {"FIXED_WORLDS_PATH": "/tmp/avalon_worlds/public_test/"}
        else:
            # Todo: private challenge phase will need to be handled differently, since the data is not
            # available inside the container.
            raise Exception("Invalid challenge phase")

        container_logs = docker_client.containers.run(
            submitted_image,
            device_requests=[docker.types.DeviceRequest(count=-1, capabilities=[["compute", "utility", "graphics"]])],
            environment=container_environment,
        )

        logger.info(f"Finished running submission. Output: {container_logs}")

        # Parse submission score from container stdout
        regex_result = re.search(r"TEST RESULTS\\n({.*})", str(container_logs))
        if regex_result is None:
            raise Exception("Submission did not output any test results")

        result_json = regex_result.group(1)
        result_json = json.loads(result_json.replace("'", '"'))

        submission_data = [
            {
                "split": "default_split",
                "show_to_participant": True,
                "accuracies": {"Total": result_json["overall_success_rate"]},
            }
        ]

        # Update EvalAI after calculating final set of metrics and set submission status as "FINISHED"
        evalai.update_submission_data(
            {
                "challenge_phase": submission["challenge_phase"],
                "submission": submission["id"],
                "stdout": container_logs,
                "submission_status": "FINISHED",
                "result": json.dumps(submission_data),
            }
        )

        logger.info(f"Reported {submission['id']} to EvalAI: {submission_data}")

    except Exception as e:
        # Update EvalAI in case of errors and set submission status as "FAILED"
        evalai.update_submission_data(
            {
                "challenge_phase": submission["challenge_phase"],
                "submission": submission["id"],
                "submission_status": "FAILED",
            }
        )

        logger.error(f"Failed to run submitted image: {e}")
        raise


if __name__ == "__main__":
    auth_token = os.getenv("EVALAI_AUTH_TOKEN")
    evalai_api_server = "https://eval.ai"
    queue_name = "random-number-generator-challenge-1882-production-3364fa36-c6a1-47b5-a7b4-2e6400"
    challenge_pk = "1882"

    evalai = EvalAI_Interface(auth_token, evalai_api_server, queue_name, challenge_pk)

    while True:
        # Get the message from the queue
        message = evalai.get_message_from_sqs_queue()
        message_body = message.get("body")
        if message_body:
            submission_pk = message_body.get("submission_pk")
            challenge_pk = message_body.get("challenge_pk")
            phase_pk = message_body.get("phase_pk")
            # Get submission details -- This will contain the input file URL
            submission: Dict[str, Any] = evalai.get_submission_by_pk(submission_pk)

            if (
                submission.get("status") == "finished"
                or submission.get("status") == "failed"
                or submission.get("status") == "cancelled"
            ):
                message_receipt_handle = message.get("receipt_handle")
                evalai.delete_message_from_sqs_queue(message_receipt_handle)

            elif submission.get("status") == "running":
                # Do nothing on EvalAI
                pass

            else:
                handle_submission(evalai, submission)

        # Poll challenge queue for new submissions
        logger.info("Waiting for submission...")
        time.sleep(30)



import json
from collections import defaultdict
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Protocol
from typing import Set
from typing import Tuple

import attr
import numpy as np
from godot_parser import GDObject
from godot_parser import Node as GDNode
from godot_parser import Vector3 as GDVector3
from scipy.spatial.transform import Rotation

from avalon.common.utils import float_to_str
from avalon.common.utils import only
from avalon.datagen.godot_base_types import FloatRange
from avalon.datagen.world_creation.configs.flora import FloraConfig
from avalon.datagen.world_creation.constants import IS_ALL_SCENERY_IN_MAIN
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.constants import WATER_LINE
from avalon.datagen.world_creation.debug_plots import plot_value_grid_multi_marker
from avalon.datagen.world_creation.entities.animals import Predator
from avalon.datagen.world_creation.entities.animals import Prey
from avalon.datagen.world_creation.entities.entity import Entity
from avalon.datagen.world_creation.entities.food import Food
from avalon.datagen.world_creation.entities.scenery import Scenery
from avalon.datagen.world_creation.entities.spawn_point import PLAYER_SPAWN_POINT
from avalon.datagen.world_creation.entities.spawn_point import SpawnPoint
from avalon.datagen.world_creation.entities.tools.boulder import Boulder
from avalon.datagen.world_creation.entities.tools.colored_sphere import ColoredSphere
from avalon.datagen.world_creation.entities.tools.log import Log
from avalon.datagen.world_creation.entities.tools.stone import Stone
from avalon.datagen.world_creation.entities.tools.weapons import Weapon
from avalon.datagen.world_creation.region import Region
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.biome_map import make_fast_biome
from avalon.datagen.world_creation.worlds.height_map import clamp
from avalon.datagen.world_creation.worlds.height_map import get_flora_config_by_file
from avalon.datagen.world_creation.worlds.terrain import Terrain
from avalon.datagen.world_creation.worlds.terrain import _create_static_body
from avalon.datagen.world_creation.worlds.terrain import create_multimesh_instance
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world import get_spawn


def get_world_slug(
    task_name: str, seed: float, difficulty: float, is_practice: bool = False, index: Optional[int] = None
) -> str:
    index_part = f"_{index}" if index is not None else ""
    slug = f"{task_name}__{seed}{index_part}__{float_to_str(difficulty)}"
    if is_practice:
        return f"practice__{slug}"
    return slug


# rand is required because technically we generate the biome map and terrain first, then export
# if you'd like to non-randomly export, simply call export_world_terrain below
def export_world(
    output_folder: Path,
    rand: np.random.Generator,
    world: World,
) -> None:
    # first we create the biome map
    if world.export_config.is_biome_fast:
        biome_config = attr.evolve(world.biome_config, is_scenery_added=False)
        biome_map = make_fast_biome(world.map, biome_config)
    else:
        biome_map, world = world.make_natural_biomes(rand)

    # then we use that to create the terrain
    terrain = world.generate_terrain(rand, biome_map)

    # then we can actually do the export
    export_world_terrain(output_folder, world, terrain)


class HasPosition(Protocol):
    position: np.ndarray


def export_world_terrain(
    output_folder: Path,
    world: World,
    terrain: Terrain,
):
    output_folder.mkdir(parents=True, exist_ok=True)

    if world.export_config.debug_visualization_config is not None:
        world = _create_visualization_world(output_folder, world)

    is_exported_with_absolute_paths = world.export_config.is_exported_with_absolute_paths
    biome_map = terrain.biome_map
    map = biome_map.map
    building_name_by_id = {}
    for i, building in world.building_by_id.items():
        building_name = f"building_{i}"
        building_name_by_id[i] = building_name
    start_tile: Optional[Path] = None
    assert map.region.x.size == map.region.z.size

    # safety check
    items_outside_world = [x for x in world.items if not map.region.contains_point_2d(x.point2d)]
    if len(items_outside_world) > 0:
        raise Exception("Some items were spawned outside of the world, should never happen")

    # our tiling logic is basically--only bother exporting with tiles if the world is sufficiently large and it
    # is enabled
    if world.config.size_in_meters > 100.0 and world.export_config.is_tiled:
        # this number controls how many "tiles" will be exported for this world. Serves as a simple LOD (level of
        # detail) parameter that allows us to run much larger worlds even on low-resource devices like the oculus
        # it has been set heuristically to this value which is a fine trade-off for export time vs runtime
        tile_count = 7
    else:
        tile_count = 1

    # how many tiles away to load when running. This only really applies to the case where tile_count > 0
    # we need to know this here so that we can decide which tiles get lumped into the "distant" tile, which is
    # exported at a lower level of detail
    tile_radius = 1

    x_tile_count = tile_count
    z_tile_count = tile_count
    x_tile_size = map.region.x.size / x_tile_count
    z_tile_size = map.region.x.size / z_tile_count
    x_min = map.region.x.min_ge
    z_min = map.region.z.min_ge

    items = []
    items_by_tile_ids: Dict[Tuple[int, int], List[Entity]] = {}
    trees_by_tile_ids_and_resource: Dict[Tuple[int, int], Dict[str, List[Scenery]]] = {}
    flora_by_tile_ids_and_resource: Dict[Tuple[int, int], Dict[str, List[Scenery]]] = {}
    for item in world.items:
        zx_tile_ids = _item_to_tile_ids(item, world.map.region, x_tile_count, z_tile_count)
        if isinstance(item, Scenery):
            scenery_by_tile_ids_and_resource = (
                trees_by_tile_ids_and_resource if "trees/" in item.resource_file else flora_by_tile_ids_and_resource
            )
            if zx_tile_ids not in scenery_by_tile_ids_and_resource:
                scenery_by_tile_ids_and_resource[zx_tile_ids] = {}
            resource = item.resource_file
            if resource not in scenery_by_tile_ids_and_resource[zx_tile_ids]:
                scenery_by_tile_ids_and_resource[zx_tile_ids][resource] = []
            scenery_by_tile_ids_and_resource[zx_tile_ids][resource].append(item)
        else:
            if zx_tile_ids not in items_by_tile_ids:
                items_by_tile_ids[zx_tile_ids] = []
            items_by_tile_ids[zx_tile_ids].append(item)
            items.append(item)

    if world.export_config.is_minor_scenery_hidden:
        flora_by_tile_ids_and_resource = {}

    all_tile_ids = set()
    for x_idx in range(0, x_tile_count):
        for z_idx in range(0, z_tile_count):
            all_tile_ids.add((z_idx, x_idx))

    terrain_export_results = []
    spawn_point = None
    spawn_point_names: Set[str] = set()
    for x_idx in range(0, x_tile_count):
        for z_idx in range(0, z_tile_count):
            region = Region(
                x=FloatRange(x_min + x_idx * x_tile_size, x_min + (x_idx + 1) * x_tile_size),
                z=FloatRange(z_min + z_idx * z_tile_size, z_min + (z_idx + 1) * z_tile_size),
            )
            region_items = items_by_tile_ids.get((z_idx, x_idx), [])
            region_building_names = [
                building_name_by_id[i] for i, x in world.building_by_id.items() if region.overlaps_region(x.region)
            ]
            fine_output_file = output_folder / f"tile_{x_idx}_{z_idx}.tscn"
            coarse_output_file = output_folder / f"distant_{x_idx}_{z_idx}.tscn"
            # extend the outside regions a tiny bit so that no triangles are left behind.
            adapted_region = region.epsilon_expand(
                x_idx == 0,
                x_idx == x_tile_count - 1,
                z_idx == 0,
                z_idx == z_tile_count - 1,
            )
            neighboring_and_this_tile_ids = set()
            for i in range(-tile_radius, tile_radius + 1):
                other_tile_x = x_idx + i
                if other_tile_x < 0 or other_tile_x >= x_tile_count:
                    continue
                for j in range(-tile_radius, tile_radius + 1):
                    other_tile_z = z_idx + j
                    if other_tile_z < 0 or other_tile_z >= z_tile_count:
                        continue
                    neighboring_and_this_tile_ids.add((other_tile_z, other_tile_x))
            distant_tile_ids = set([x for x in all_tile_ids if x not in neighboring_and_this_tile_ids])
            terrain_export_results.append(
                terrain.export(
                    fine_output_file,
                    coarse_output_file,
                    adapted_region,
                    x_idx,
                    z_idx,
                    tile_radius,
                    region_building_names,
                    trees_by_tile_ids_and_resource if not IS_ALL_SCENERY_IN_MAIN else {},
                    flora_by_tile_ids_and_resource if not IS_ALL_SCENERY_IN_MAIN else {},
                    distant_tile_ids,
                    neighboring_and_this_tile_ids,
                )
            )
            spawn_items = [x for x in region_items if isinstance(x, SpawnPoint)]
            for spawn in spawn_items:
                assert spawn.name not in spawn_point_names, f"SpawnPoints name {spawn.name} is duplicated"
                spawn_point_names.add(spawn.name)
                if spawn.name == PLAYER_SPAWN_POINT:
                    assert spawn_point is None, "Should only be one player spawn point"
                    spawn_point = spawn
                    start_tile = fine_output_file
    assert start_tile is not None, "No tiles contained a spawn point--that's not good!"

    starting_terrain_tile_path = (
        str(output_folder / start_tile.name) if is_exported_with_absolute_paths else "./" + start_tile.name
    )

    # export tscn file
    scene = GodotScene()
    level_resource = scene.add_ext_resource(starting_terrain_tile_path, "PackedScene")
    terrain_manager_script = scene.add_ext_resource("res://terrain/terrain_manager.gd", "Script")

    if world.export_config.is_meta_data_exported:
        meta_output_path = output_folder / "meta.json"
        with open(meta_output_path, "w") as outfile:
            # spawn_point
            assert spawn_point is not None
            food: List[HasPosition] = [x for x in items if isinstance(x, Food)]
            prey: List[HasPosition] = [x for x in items if isinstance(x, Prey)]
            food_count = len(food) + len(prey)
            if food_count == 1:
                only_food = only(food + prey)
                total_distance = np.linalg.norm(only_food.position - spawn_point.position)
            else:
                all_food = food + prey
                total_distance = np.float64(sum([np.linalg.norm(x.position - spawn_point.position) for x in all_food]))
            meta_data = dict(
                size_in_meters=world.config.size_in_meters,
                food_count=food_count,
                total_distance=total_distance,
                is_visibility_required=spawn_point.is_visibility_required,
            )
            outfile.write(json.dumps(meta_data))

    main_output_path = output_folder / "main.tscn"

    with scene.use_tree() as tree:
        tree.root = GDNode("Avalon", type="Spatial")
        if biome_map.config.export_config is None or biome_map.config.export_config.is_sun_enabled:
            tree.root.add_child(
                GDNode(
                    "Sun",
                    type="DirectionalLight",
                    properties={
                        "transform": _get_transform_from_pitch_and_yaw(
                            *_get_pitch_and_yaw_from_sky_params(biome_map.config.godot_sky_config)
                        ),
                        **biome_map.config.godot_sun_config,
                    },
                )
            )
        camera_transform = GDObject(
            "Transform", 0.5, 0.594, -0.629, 0, 0.727, 0.686, 0.866, -0.343, 0.363, -6.525, 5.377, 4.038
        )
        tree.root.add_child(
            GDNode(
                "Camera",
                type="Camera",
                properties={"transform": camera_transform},
            )
        )
        tree.root.add_child(
            _create_godot_world_environment(
                scene, biome_map.config.godot_env_config, biome_map.config.godot_sky_config
            )
        )

        if IS_ALL_SCENERY_IN_MAIN:
            scenery_by_resource: Dict[str, List[Scenery]] = {}
            for scenery_by_tile_ids_and_resource in (
                trees_by_tile_ids_and_resource,
                flora_by_tile_ids_and_resource,
            ):
                for tile_id, inner_dict in scenery_by_tile_ids_and_resource.items():
                    for resource, scenery_list in inner_dict.items():
                        if resource not in scenery_by_resource:
                            scenery_by_resource[resource] = []
                        scenery_by_resource[resource].extend(scenery_list)
                for resource, scenery_list in scenery_by_resource.items():
                    tree.root.add_child(
                        create_multimesh_instance(
                            scene,
                            "main",
                            scenery_list,
                            resource.replace(".tscn", ".res"),
                            biome_map.config.flora_config,
                        )
                    )

        data = np.logical_not(world.is_climbable)
        if data.shape[1] % 8 != 0:
            data = np.pad(data, pad_width=[(0, 0), (0, 8 - data.shape[1] % 8)])
        bytes = np.packbits(data, bitorder="little")

        boundary = world.config.size_in_meters / 2.0

        export_path = "."
        if is_exported_with_absolute_paths:
            # TODO: fix the double quotes, update S3 checksum manifest
            export_path = f'"{str(output_folder)}"'
        if world.export_config.world_id is not None:
            export_path = f"res://worlds/{world.export_config.world_id}"

        terrain_manager = GDNode(
            "TerrainManager",
            type="Node",
            properties=dict(
                script=terrain_manager_script.reference,
                export_path=export_path,
                x_tile_count=x_tile_count,
                z_tile_count=z_tile_count,
                tile_radius=tile_radius,
                x_min=-boundary,
                z_min=-boundary,
                x_max=boundary,
                z_max=boundary,
                climb_map=GDObject("PoolByteArray", *list(bytes)),
                climb_map_x=len(world.is_climbable[0, :]),
                climb_map_y=len(world.is_climbable[:, 0]),
            ),
        )
        tiles_node = GDNode("tiles", type="Node")
        tiles_node.add_child(GDNode(str(start_tile.name).split(".")[0], instance=level_resource.reference.id))
        terrain_manager.add_child(tiles_node)

        building_group = GDNode("buildings", type="Node")
        terrain_manager.add_child(building_group)
        for i, building in world.building_by_id.items():
            building.export(
                scene,
                building_group,
                building_name_by_id[i],
                is_indoor_lighting_enabled=world.export_config.is_indoor_lighting_enabled,
            )

        terrain_manager.add_child(GDNode("walls", type="Node"))
        collision_mesh_group = GDNode("terrain_collision_meshes", type="Node")

        for i, (region_triangles, region_vertices) in enumerate(terrain_export_results):
            collision_mesh_group.add_child(_create_static_body(i, scene, region_triangles, region_vertices))
        terrain_manager.add_child(collision_mesh_group)
        tree.root.add_child(terrain_manager)

        tracker_node = GDNode("dynamic_tracker", type="Node")
        next_id = max([x.entity_id for x in items]) + 1
        for i, item in enumerate(items):
            if item.entity_id < 0:
                item = attr.evolve(item, entity_id=next_id)
                next_id += 1
            tracker_node.add_child(item.get_node(scene))
        tree.root.add_child(tracker_node)

        # figure out all of the trees
        tree_by_resource: Dict[str, List[Scenery]] = {}
        for tile_id, inner_dict in trees_by_tile_ids_and_resource.items():
            for resource, scenery_list in inner_dict.items():
                if resource not in tree_by_resource:
                    tree_by_resource[resource] = []
                tree_by_resource[resource].extend(scenery_list)

        # add all collision shapes for the trees
        tree_collision_mesh_group = GDNode("tree_collision_meshes", type="Spatial")
        for i, (resource, scenery_list) in enumerate(tree_by_resource.items()):
            for j, tree_obj in enumerate(scenery_list):
                tree_collider = _create_tree_shape(scene, tree_obj, i, j, world.biome_config.flora_config, resource)
                tree_collision_mesh_group.add_child(tree_collider)
        tree.root.add_child(tree_collision_mesh_group)

        # is big enough that you dont really notice that it isn't infinite
        tree.root.add_child(_create_ocean_node(scene, "large_ocean", ocean_size=1_000, offset=-2))

        # if we ONLY make a huge ocean, it has a rendering bug near the shore
        # so we also make a small one that is slightly higher
        # UNLESS it is a completely flat world (probably indoor-only), since we'd get overlapping planes
        world_size = round(max([map.region.x.size, map.region.z.size]) * 1.2)
        if not (map.Z == 0).all():
            tree.root.add_child(_create_ocean_node(scene, "near_ocean", ocean_size=world_size, offset=-1))

        tree.root.add_child(
            _create_ocean_collision(
                scene, "ocean", position=np.array([0, -1.5, 0]), size=np.array([world_size, 1.0, world_size])
            )
        )

    scene.write(str((main_output_path).absolute()))


def _create_visualization_world(output_folder: Path, world: World) -> World:
    spawn_item = only([x for x in world.items if isinstance(x, SpawnPoint)])
    food_items: List[HasPosition] = [x for x in world.items if isinstance(x, Food)]
    if len(food_items) == 0:
        food_items = [x for x in world.items if isinstance(x, Prey)]
    spawn_position = spawn_item.position
    food_distances = [(np.linalg.norm(spawn_position - x.position), x) for x in food_items]
    food_position = sorted(food_distances)[-1][1].position
    mid_position = (spawn_position + food_position) / 2.0
    spawn_to_goal_vec = food_position - spawn_position
    new_spawn_item = None
    for off_center in (1.0, 0.5, 0.0):
        for orientation in (1.0, -1.0):
            sideways = np.array([-spawn_to_goal_vec[2], 0.0, spawn_to_goal_vec[0]]) * orientation * off_center
            full_len = np.linalg.norm(spawn_to_goal_vec)
            new_spawn = mid_position + sideways + UP_VECTOR * full_len / 2.0
            if not world.map.region.contains_point_3d(new_spawn, epsilon=0.0):
                continue
            rand = np.random.default_rng(0)
            new_spawn_item = get_spawn(rand, 0.0, new_spawn, mid_position)
            # markers = [
            #     world.map.point_to_index(to_2d_point(x))
            #     for x in [food_position, mid_position, spawn_position, new_spawn_item.position]
            # ]
            # plot_value_grid(world.map.Z, markers=markers)
            break
        if new_spawn_item is not None:
            break
    assert new_spawn_item is not None, "Failed to find a point in the region to view from"
    world_items = [x for x in world.items if not isinstance(x, SpawnPoint)]
    world_items.append(new_spawn_item)

    # replace all items with markers
    points_by_type = defaultdict(list)
    new_items = []
    collected_types = tuple([Food, Prey, Predator, Log, Weapon, Boulder, Stone])
    for item in world_items:
        if isinstance(item, collected_types):
            for dtype in collected_types:
                if isinstance(item, dtype):
                    points_by_type[dtype.__name__].append(item.position)
        else:
            new_items.append(item)
    SPAWN_COLOR = "#FF00F8"
    color_by_type_name = {
        Food.__name__: "#FF0000",
        Prey.__name__: "#FDFF00",
        Predator.__name__: "#AE00FF",
        Log.__name__: "#FF9E00",
        Weapon.__name__: "#000000",
        Boulder.__name__: "#FF0094",
        Stone.__name__: "#FF0094",
    }
    for type_name, points in points_by_type.items():
        color = color_by_type_name[type_name]
        for point in points:
            new_items.append(ColoredSphere(position=point, color=color))
    new_items.append(ColoredSphere(position=spawn_position, color=SPAWN_COLOR))
    world = attr.evolve(world, items=tuple(new_items))

    if (
        world.export_config.debug_visualization_config is not None
        and world.export_config.debug_visualization_config.is_2d_graph_drawn
    ):
        marker_lists = []
        for type_name, points in points_by_type.items():
            color = color_by_type_name[type_name]
            markers = [world.map.point_to_index(to_2d_point(x)) for x in points]
            marker_lists.append((markers, color))
        marker_lists.append(([world.map.point_to_index(to_2d_point(spawn_position))], SPAWN_COLOR))
        height_with_water = world.map.Z.copy()
        height_with_water[height_with_water < WATER_LINE] = world.map.Z.max(initial=1.0)
        tag = ""
        if world.config.is_indoor_only:
            tag = "INDOOR"
            height_with_water[:, :] = 0.5
        plot_value_grid_multi_marker(height_with_water, tag + str(output_folder), marker_lists)

    return world


def _create_ocean_node(scene: GodotScene, name: str, ocean_size: int, offset: float):
    ocean_location = np.array([0, offset, 0])
    ocean_node = GDNode(
        name,
        type="MeshInstance",
        properties={
            "transform": GDObject("Transform", 1, 0, 0, 0, 1, 0, 0, 0, 1, *ocean_location),
            "mesh": scene.add_sub_resource(
                "CubeMesh",
                material=scene.add_ext_resource("res://materials/ocean.material", "Material").reference,
                size=GDVector3(ocean_size, 2.0, ocean_size),
            ).reference,
            "material/0": None,
        },
    )
    return ocean_node


def _item_to_tile_ids(item: Entity, world_region: Region, x_tile_count: int, z_tile_count: int) -> Tuple[int, int]:
    if x_tile_count == 1:
        assert z_tile_count == 1, "lazy implementation"
        return (0, 0)
    max_x_index = x_tile_count - 1
    one_cell_x = world_region.x.size / (max_x_index)
    x = round((item.position[0] - world_region.x.min_ge) / one_cell_x)
    max_y_index = z_tile_count - 1
    one_cell_y = world_region.z.size / (max_y_index)
    y = round((item.position[2] - world_region.z.min_ge) / one_cell_y)
    return clamp(y, 0, max_y_index), clamp(x, 0, max_x_index)


def _get_pitch_and_yaw_from_sky_params(sky_config: Dict[str, Any]) -> Tuple[float, float]:
    pitch = (180.0 - sky_config.get("sun_latitude", 170.0)) * -1.0
    yaw = sky_config.get("sun_longitude", 0.0) * -1.0
    return pitch, yaw


def _get_transform_from_pitch_and_yaw(pitch: float, yaw: float) -> GDObject:
    yaw_rotation = Rotation.from_euler("y", yaw, degrees=True)
    pitch_rotation = Rotation.from_euler("x", pitch, degrees=True)
    rotation = (yaw_rotation * pitch_rotation).as_matrix().flatten()
    return GDObject("Transform", *rotation, 0, 0, 0)


def _create_godot_world_environment(
    scene: GodotScene, env_config: Dict[str, Any], sky_config: Dict[str, Any]
) -> GDNode:
    sky_resource = scene.add_sub_resource("ProceduralSky", **sky_config)
    full_env_config = {**env_config}
    full_env_config["background_sky"] = sky_resource.reference
    env_resource = scene.add_sub_resource("Environment", **full_env_config)
    return GDNode(
        "WorldEnvironment",
        type="WorldEnvironment",
        properties={
            "environment": env_resource.reference,
        },
    )


def _create_ocean_collision(scene: GodotScene, name: str, position: np.ndarray, size: np.ndarray) -> GDNode:
    transform = GDObject("Transform", 1, 0, 0, 0, 1, 0, 0, 0, 1, *position)
    root = GDNode(
        name,
        type="StaticBody",
        properties={
            "transform": transform,
        },
    )
    root.add_child(
        GDNode(
            f"{name}_collision",
            type="CollisionShape",
            properties={
                "shape": scene.add_sub_resource(
                    "BoxShape", resource_local_to_scene=True, extents=GDVector3(*size / 2.0)
                ).reference
            },
        )
    )

    return root


def _create_tree_shape(
    scene: GodotScene, tree_obj: Scenery, i: int, j: int, flora_config: Dict[str, FloraConfig], resource: str
) -> GDNode:
    config, resource_name = get_flora_config_by_file(flora_config, resource)
    assert config is not None
    size = config.collision_extents * tree_obj.scale * config.default_scale
    offset_position = tree_obj.position.copy()
    offset_position[1] += config.collision_extents[1] * tree_obj.scale[1] * config.default_scale
    transform = GDObject("Transform", *tree_obj.rotation, *offset_position)
    tree_collider = GDNode(f"tree_{i}_collision_{j}", type="StaticBody", properties={"transform": transform})
    tree_collider.add_child(
        GDNode(
            f"tree_{i}_collision_{j}_shape",
            type="CollisionShape",
            properties={
                "shape": scene.add_sub_resource(
                    "BoxShape", resource_local_to_scene=True, extents=GDVector3(*size)
                ).reference
            },
        )
    )

    return tree_collider
